import os

from datetime import datetime
from typing import Dict, Optional

from polyaxon._compiler.contexts.dask_job import DaskJobContextsManager
from polyaxon._compiler.contexts.job import JobContextsManager
from polyaxon._compiler.contexts.kubeflow import (
    MPIJobContextsManager,
    MXJobContextsManager,
    PaddleJobContextsManager,
    PytorchJobContextsManager,
    TfJobContextsManager,
    XGBoostJobContextsManager,
)
from polyaxon._compiler.contexts.ray_job import RayJobContextsManager
from polyaxon._compiler.contexts.service import ServiceContextsManager
from polyaxon._connections import V1Connection
from polyaxon._contexts import keys as ctx_keys
from polyaxon._contexts import paths as ctx_paths
from polyaxon._contexts import sections as ctx_sections
from polyaxon._flow import V1CloningKind, V1CompiledOperation, V1Plugins, V1RunKind
from polyaxon._utils.fqn_utils import get_project_instance, get_run_instance
from polyaxon.exceptions import PolyaxonCompilerError

CONTEXTS_MANAGERS = {
    V1RunKind.CLEANER: JobContextsManager,
    V1RunKind.NOTIFIER: JobContextsManager,
    V1RunKind.TUNER: JobContextsManager,
    V1RunKind.WATCHDOG: JobContextsManager,
    V1RunKind.JOB: JobContextsManager,
    V1RunKind.SERVICE: ServiceContextsManager,
    V1RunKind.MPIJOB: MPIJobContextsManager,
    V1RunKind.TFJOB: TfJobContextsManager,
    V1RunKind.MXJOB: MXJobContextsManager,
    V1RunKind.XGBJOB: XGBoostJobContextsManager,
    V1RunKind.PYTORCHJOB: PytorchJobContextsManager,
    V1RunKind.PADDLEJOB: PaddleJobContextsManager,
    V1RunKind.RAYJOB: RayJobContextsManager,
    V1RunKind.DASKJOB: DaskJobContextsManager,
}


def resolve_globals_contexts(
    namespace: str,
    owner_name: str,
    project_name: str,
    project_uuid: str,
    run_uuid: str,
    run_name: str,
    run_path: str,
    iteration: int,
    created_at: datetime,
    compiled_at: datetime,
    schedule_at: Optional[datetime] = None,
    started_at: Optional[datetime] = None,
    finished_at: Optional[datetime] = None,
    duration: Optional[float] = None,
    plugins: Optional[V1Plugins] = None,
    artifacts_store: V1Connection = None,
    cloning_kind: V1CloningKind = None,
    original_uuid: Optional[str] = None,
    is_independent: bool = True,
) -> Dict:
    resolved_contexts = {
        ctx_sections.GLOBALS: {
            ctx_keys.OWNER_NAME: owner_name,
            ctx_keys.PROJECT_NAME: project_name,
            ctx_keys.PROJECT_UNIQUE_NAME: get_project_instance(
                owner_name, project_name
            ),
            ctx_keys.PROJECT_UUID: project_uuid,
            ctx_keys.RUN_INFO: get_run_instance(owner_name, project_name, run_uuid),
            ctx_keys.NAME: run_name,
            ctx_keys.UUID: run_uuid,
            ctx_keys.NAMESPACE: namespace,
            ctx_keys.ITERATION: iteration,
            ctx_keys.CONTEXT_PATH: ctx_paths.CONTEXT_ROOT,
            ctx_keys.ARTIFACTS_PATH: ctx_paths.CONTEXT_MOUNT_ARTIFACTS,
            ctx_keys.CREATED_AT: created_at,
            ctx_keys.COMPILED_AT: compiled_at,
            ctx_keys.SCHEDULE_AT: schedule_at,
            ctx_keys.STARTED_AT: started_at,
            ctx_keys.FINISHED_AT: finished_at,
            ctx_keys.DURATION: duration,
            ctx_keys.CLONING_KIND: cloning_kind,
            ctx_keys.ORIGINAL_UUID: original_uuid,
            ctx_keys.IS_INDEPENDENT: is_independent,
            ctx_keys.STORE_PATH: "",
        },
    }

    plugins = V1Plugins.get_or_create(plugins)
    if plugins.collect_artifacts:
        run_artifacts_path = ctx_paths.CONTEXT_MOUNT_ARTIFACTS_FORMAT.format(run_path)
        run_outputs_path = ctx_paths.CONTEXT_MOUNT_RUN_OUTPUTS_FORMAT.format(run_path)
        resolved_contexts[ctx_sections.GLOBALS][
            ctx_keys.RUN_ARTIFACTS_PATH
        ] = run_artifacts_path
        resolved_contexts[ctx_sections.GLOBALS][
            ctx_keys.RUN_OUTPUTS_PATH
        ] = run_outputs_path
    elif artifacts_store:
        run_artifacts_path = os.path.join(artifacts_store.store_path, run_path)
        run_outputs_path = os.path.join(run_artifacts_path, "outputs")
        resolved_contexts[ctx_sections.GLOBALS][
            ctx_keys.RUN_ARTIFACTS_PATH
        ] = run_artifacts_path
        resolved_contexts[ctx_sections.GLOBALS][
            ctx_keys.RUN_OUTPUTS_PATH
        ] = run_outputs_path

    if plugins.mount_artifacts_store and artifacts_store:
        resolved_contexts[ctx_sections.GLOBALS][
            ctx_keys.STORE_PATH
        ] = artifacts_store.store_path
    return resolved_contexts


def resolve_contexts(
    namespace: str,
    owner_name: str,
    project_name: str,
    project_uuid: str,
    run_uuid: str,
    run_name: str,
    run_path: str,
    compiled_operation: V1CompiledOperation,
    artifacts_store: V1Connection,
    connection_by_names: Dict[str, V1Connection],
    iteration: int,
    created_at: datetime,
    compiled_at: datetime,
    schedule_at: Optional[datetime] = None,
    started_at: Optional[datetime] = None,
    finished_at: Optional[datetime] = None,
    duration: Optional[float] = None,
    cloning_kind: V1CloningKind = None,
    original_uuid: Optional[str] = None,
    is_independent: bool = True,
) -> Dict:
    run_kind = compiled_operation.get_run_kind()
    if run_kind not in CONTEXTS_MANAGERS:
        raise PolyaxonCompilerError(
            "Contexts manager Error. "
            "Specification with run kind: {} is not supported in this deployment version".format(
                run_kind
            )
        )

    resolved_contexts = resolve_globals_contexts(
        namespace=namespace,
        owner_name=owner_name,
        project_name=project_name,
        project_uuid=project_uuid,
        run_uuid=run_uuid,
        run_name=run_name,
        run_path=run_path,
        iteration=iteration,
        created_at=created_at,
        compiled_at=compiled_at,
        schedule_at=schedule_at,
        started_at=started_at,
        finished_at=finished_at,
        duration=duration,
        plugins=compiled_operation.plugins,
        artifacts_store=artifacts_store,
        cloning_kind=cloning_kind,
        original_uuid=original_uuid,
        is_independent=is_independent,
    )

    return CONTEXTS_MANAGERS[run_kind].resolve(
        namespace=namespace,
        owner_name=owner_name,
        project_name=project_name,
        run_uuid=run_uuid,
        contexts=resolved_contexts,
        compiled_operation=compiled_operation,
        connection_by_names=connection_by_names,
    )



from typing import Dict, List, Union

from polyaxon._docker import docker_types
from polyaxon._k8s import k8s_schemas

EnvVar = Union[k8s_schemas.V1EnvVar, docker_types.V1EnvVar]
ResourceRequirements = Union[
    k8s_schemas.V1ResourceRequirements, docker_types.V1ResourceRequirements
]
Container = Union[k8s_schemas.V1Container, docker_types.V1Container]
ContainerPort = Union[k8s_schemas.V1ContainerPort, docker_types.V1ContainerPort]
Resource = Union[Dict, List[docker_types.V1Container]]
VolumeMount = Union[k8s_schemas.V1VolumeMount, docker_types.V1VolumeMount]



from typing import List, Optional

from polyaxon._connections import V1Connection, V1ConnectionResource
from polyaxon._contexts import paths as ctx_paths
from polyaxon._k8s import k8s_schemas
from polyaxon._runner.converter import BaseConverter
from polyaxon._runner.converter.common import constants


class MountsMixin(BaseConverter):
    @classmethod
    def _get_mount_from_store(
        cls, store: V1Connection
    ) -> Optional[k8s_schemas.V1VolumeMount]:
        if not store or not store.is_mount:
            return None

        return k8s_schemas.V1VolumeMount(
            name=store.name,
            mount_path=store.schema_.mount_path,
            read_only=store.schema_.read_only,
        )

    @staticmethod
    def _get_mount_from_resource(
        resource: V1ConnectionResource,
    ) -> Optional[k8s_schemas.V1VolumeMount]:
        if not resource or not resource.mount_path:
            return None

        return k8s_schemas.V1VolumeMount(
            name=resource.name, mount_path=resource.mount_path, read_only=True
        )

    @classmethod
    def _get_docker_context_mount(cls) -> k8s_schemas.V1VolumeMount:
        return k8s_schemas.V1VolumeMount(
            name=constants.VOLUME_MOUNT_DOCKER,
            mount_path=ctx_paths.CONTEXT_MOUNT_DOCKER,
        )

    @staticmethod
    def _get_auth_context_mount(
        read_only: Optional[bool] = None,
        run_path: Optional[str] = None,
    ) -> k8s_schemas.V1VolumeMount:
        return k8s_schemas.V1VolumeMount(
            name=constants.VOLUME_MOUNT_CONFIGS,
            mount_path=ctx_paths.CONTEXT_MOUNT_CONFIGS,
            read_only=read_only,
        )

    @staticmethod
    def _get_artifacts_context_mount(
        read_only: Optional[bool] = None,
        run_path: Optional[str] = None,
    ) -> k8s_schemas.V1VolumeMount:
        return k8s_schemas.V1VolumeMount(
            name=constants.VOLUME_MOUNT_ARTIFACTS,
            mount_path=ctx_paths.CONTEXT_MOUNT_ARTIFACTS,
            read_only=read_only,
        )

    @staticmethod
    def _get_connections_context_mount(
        name: str, mount_path: str, run_path: str
    ) -> k8s_schemas.V1VolumeMount:
        return k8s_schemas.V1VolumeMount(name=name, mount_path=mount_path)

    @staticmethod
    def _get_shm_context_mount() -> k8s_schemas.V1VolumeMount:
        """
        Mount an tmpfs volume to /dev/shm.
        This will set /dev/shm size to half of the RAM of node.
        By default, /dev/shm is very small, only 64MB.
        Some experiments will fail due to lack of share memory,
        such as some experiments running on Pytorch.
        """
        return k8s_schemas.V1VolumeMount(
            name=constants.VOLUME_MOUNT_SHM, mount_path=ctx_paths.CONTEXT_MOUNT_SHM
        )

    @classmethod
    def _get_mounts(
        cls,
        use_auth_context: bool,
        use_docker_context: bool,
        use_shm_context: bool,
        use_artifacts_context: bool,
        run_path: Optional[str] = None,
    ) -> List[k8s_schemas.V1VolumeMount]:
        mounts = []
        if use_auth_context:
            mounts.append(
                cls._get_auth_context_mount(read_only=True, run_path=run_path)
            )
        if use_artifacts_context:
            mounts.append(
                cls._get_artifacts_context_mount(read_only=False, run_path=run_path)
            )
        if use_docker_context:
            mounts.append(cls._get_docker_context_mount())
        if use_shm_context:
            mounts.append(cls._get_shm_context_mount())

        return mounts



import copy

from typing import Dict, List, Optional, Set, Type, Union

from polyaxon._flow import (
    ParamSpec,
    V1CompiledOperation,
    V1Dag,
    V1Hook,
    V1Init,
    V1Param,
    validate_run_patch,
)
from polyaxon._polyaxonfile.specs import kinds
from polyaxon._polyaxonfile.specs.base import BaseSpecification
from polyaxon._polyaxonfile.specs.libs.parser import PolyaxonfileParser
from polyaxon._polyaxonfile.specs.operation import OperationSpecification
from polyaxon.exceptions import PolyaxonfileError, PolyaxonSchemaError


class CompiledOperationSpecification(BaseSpecification):
    """The polyaxonfile specification for compiled operation."""

    _SPEC_KIND = kinds.COMPILED_OPERATION

    CONFIG: Type[V1CompiledOperation] = V1CompiledOperation

    @staticmethod
    def dict_to_param_spec(
        contexts: Optional[Dict] = None,
        is_requested: bool = False,
        is_context: bool = False,
    ):
        contexts = contexts or {}
        return {
            k: ParamSpec(
                name=k,
                param=V1Param(value=v),
                type="Any",
                is_flag=False,
                is_list=None,
                is_requested=is_requested,
                is_context=is_context,
                arg_format=None,
            )
            for k, v in contexts.items()
        }

    @classmethod
    def calculate_context_spec(
        cls,
        config: V1CompiledOperation,
        contexts: Optional[Dict] = None,
        should_be_resolved: bool = False,
    ) -> Dict[str, ParamSpec]:
        param_spec = config.validate_params(
            is_template=False,
            check_all_refs=True,
            parse_values=True,
            parse_joins=not should_be_resolved,
        )
        if should_be_resolved:
            for p_spec in param_spec:
                if not p_spec.param.is_literal:
                    raise PolyaxonfileError(
                        "calculate_context_spec received a non-resolved "
                        "ref param `{}` with value `{}`".format(
                            p_spec.name, p_spec.param.to_dict()
                        )
                    )
        param_spec = {param.name: param for param in param_spec}
        param_spec.update(cls.dict_to_param_spec(contexts=contexts, is_context=True))
        return param_spec

    @classmethod
    def _apply_operation_contexts(
        cls,
        config: V1CompiledOperation,
        param_spec: Dict[str, ParamSpec] = None,
        contexts: Optional[Dict] = None,
    ) -> V1CompiledOperation:
        if not param_spec:
            param_spec = cls.calculate_context_spec(config=config, contexts=contexts)

        parsed_data = PolyaxonfileParser.parse_operation(config, param_spec or {})
        return cls.CONFIG.read(parsed_data)

    @staticmethod
    def _apply_dag_context(config: V1CompiledOperation) -> V1CompiledOperation:
        dag_run = config.run  # type: V1Dag
        dag_run.process_dag()
        dag_run.validate_dag()
        dag_run.process_components(config.inputs)
        return config

    @classmethod
    def apply_operation_contexts(
        cls,
        config: V1CompiledOperation,
        param_spec: Dict[str, ParamSpec] = None,
        contexts: Optional[Dict] = None,
    ) -> V1CompiledOperation:
        if config.is_dag_run:
            return cls._apply_dag_context(config)
        else:
            return cls._apply_operation_contexts(
                config=config, param_spec=param_spec, contexts=contexts
            )

    @classmethod
    def _apply_connections_params(
        cls,
        connections: List[str],
        init: List[V1Init],
        artifact_store: Optional[str] = None,
        param_spec: Dict[str, ParamSpec] = None,
    ):
        if connections:
            connections = PolyaxonfileParser.parse_section(
                connections, param_spec=param_spec, parse_params=True
            )
        _init = []
        if init:
            for i in init:
                if (i.artifacts or i.paths) and not i.connection:
                    i.connection = artifact_store
                if i.connection:
                    i.connection = PolyaxonfileParser.parse_section(
                        i.connection, param_spec=param_spec, parse_params=True
                    )
                # resolved_i = V1Init.from_dict(
                #     PolyaxonfileParser.parse_section(
                #         i.to_dict(), param_spec=param_spec, parse_params=True
                #     )
                # )
                _init.append(i)

        # Prepend any param that has to_init after validation
        init_params = [v.to_init() for v in param_spec.values() if v.validate_to_init()]
        init_params = [v for v in init_params if v]
        _init = init_params + _init
        return _init, connections

    @classmethod
    def _apply_distributed_run_connections_params(
        cls,
        config: V1CompiledOperation,
        artifact_store: Optional[str] = None,
        param_spec: Dict[str, ParamSpec] = None,
    ):
        def _resolve_replica(replica):
            if not replica:
                return
            init, connections = cls._apply_connections_params(
                init=replica.init,
                connections=replica.connections,
                artifact_store=artifact_store,
                param_spec=param_spec,
            )
            replica.init = init
            replica.connections = connections
            return replica

        if config.is_mpi_job_run:
            config.run.launcher = _resolve_replica(config.run.launcher)
            config.run.worker = _resolve_replica(config.run.worker)
        elif config.is_tf_job_run:
            config.run.chief = _resolve_replica(config.run.chief)
            config.run.worker = _resolve_replica(config.run.worker)
            config.run.ps = _resolve_replica(config.run.ps)
            config.run.evaluator = _resolve_replica(config.run.evaluator)
        elif config.is_pytorch_job_run or config.is_paddle_job_run:
            config.run.master = _resolve_replica(config.run.master)
            config.run.worker = _resolve_replica(config.run.worker)
        elif config.is_mx_job_run:
            config.run.scheduler = _resolve_replica(config.run.scheduler)
            config.run.worker = _resolve_replica(config.run.worker)
            config.run.server = _resolve_replica(config.run.server)
            config.run.tuner = _resolve_replica(config.run.tuner)
            config.run.tuner_tracker = _resolve_replica(config.run.tuner_tracker)
            config.run.tuner_server = _resolve_replica(config.run.tuner_server)
        elif config.is_xgb_job_run:
            config.run.master = _resolve_replica(config.run.master)
            config.run.worker = _resolve_replica(config.run.worker)
        return config

    @classmethod
    def apply_run_connections_params(
        cls,
        config: V1CompiledOperation,
        artifact_store: Optional[str] = None,
        contexts: Optional[Dict] = None,
        param_spec: Dict[str, ParamSpec] = None,
    ) -> V1CompiledOperation:
        if not param_spec:
            param_spec = cls.calculate_context_spec(config=config, contexts=contexts)
        if config.is_job_run or config.is_service_run:
            init, connections = cls._apply_connections_params(
                init=config.run.init,
                connections=config.run.connections,
                artifact_store=artifact_store,
                param_spec=param_spec,
            )
            config.run.init = init
            config.run.connections = connections
            return config
        if config.is_distributed_run:
            return cls._apply_distributed_run_connections_params(
                config=config,
                artifact_store=artifact_store,
                param_spec=param_spec,
            )
        return config

    @classmethod
    def apply_params(
        cls,
        config: V1CompiledOperation,
        params: Optional[Dict] = None,
        context: Optional[Dict] = None,
    ) -> V1CompiledOperation:
        config.apply_params(params, context)
        return config

    @classmethod
    def apply_section_contexts(
        cls,
        config: V1CompiledOperation,
        section,
        contexts: Optional[Dict] = None,
        param_spec: Dict[str, ParamSpec] = None,
    ):
        if not param_spec:
            param_spec = cls.calculate_context_spec(config=config, contexts=contexts)

        return PolyaxonfileParser.parse_section(section, param_spec)

    @classmethod
    def _apply_runtime_contexts(
        cls,
        config: V1CompiledOperation,
        contexts: Optional[Dict] = None,
        param_spec: Dict[str, ParamSpec] = None,
    ) -> V1CompiledOperation:
        if not param_spec:
            param_spec = cls.calculate_context_spec(
                config=config, contexts=contexts, should_be_resolved=True
            )
        parsed_data = PolyaxonfileParser.parse_runtime(config.to_dict(), param_spec)
        return cls.CONFIG.read(parsed_data)

    @classmethod
    def _apply_distributed_runtime_contexts(
        cls,
        config: V1CompiledOperation,
        contexts: Optional[Dict] = None,
        param_spec: Dict[str, ParamSpec] = None,
    ) -> V1CompiledOperation:
        if not param_spec:
            # Calculate the param_space once with empty contexts
            replica_param_spec = cls.calculate_context_spec(
                config=config, contexts=None, should_be_resolved=True
            )
            param_spec = {}
            for k in contexts:
                param_spec[k] = copy.copy(replica_param_spec)
                param_spec[k].update(
                    cls.dict_to_param_spec(
                        contexts=contexts[k], is_requested=True, is_context=True
                    )
                )
        parsed_data = PolyaxonfileParser.parse_distributed_runtime(
            config.to_dict(), param_spec
        )
        return cls.CONFIG.read(parsed_data)

    @classmethod
    def apply_runtime_contexts(
        cls,
        config: V1CompiledOperation,
        contexts: Optional[Dict] = None,
        param_spec: Dict[str, ParamSpec] = None,
    ) -> V1CompiledOperation:
        if config.has_pipeline:
            raise PolyaxonSchemaError(
                "This method is not allowed on this specification."
            )
        if config.is_distributed_run:
            return cls._apply_distributed_runtime_contexts(
                config=config,
                contexts=contexts,
                param_spec=param_spec,
            )
        else:
            return cls._apply_runtime_contexts(
                config=config,
                contexts=contexts,
                param_spec=param_spec,
            )

    @classmethod
    def apply_hooks_contexts(
        cls,
        config: V1CompiledOperation,
        contexts: Optional[Dict] = None,
        param_spec: Dict[str, ParamSpec] = None,
    ) -> List[V1Hook]:
        if not param_spec:
            param_spec = cls.calculate_context_spec(
                config=config, contexts=contexts, should_be_resolved=True
            )
        hooks = PolyaxonfileParser.parse_hooks(config, param_spec)
        return [V1Hook.read(hook) for hook in hooks]

    @classmethod
    def apply_preset(
        cls, config: V1CompiledOperation, preset: Union[Dict, str] = None
    ) -> V1CompiledOperation:
        if not preset:
            return config
        preset = OperationSpecification.read(
            preset, is_preset=True
        )  # type: V1Operation
        if preset.run_patch:
            config.run = config.run.patch(
                validate_run_patch(preset.run_patch, config.run.kind),
                strategy=preset.patch_strategy,
            )
        patch_keys = {
            "name",
            "description",
            "tags",
            "is_approved",
            "presets",
            "queue",
            "cache",
            "build",
            "hooks",
            "events",
            "plugins",
            "termination",
            "matrix",
            "joins",
            "schedule",
            "dependencies",
            "trigger",
            "conditions",
            "skip_on_upstream_skip",
        }
        patch_keys = patch_keys.intersection(preset.__fields_set__)
        patch_data = {k: getattr(preset, k) for k in patch_keys}
        patch_compiled = V1CompiledOperation.construct(**patch_data)
        return config.patch(patch_compiled, strategy=preset.patch_strategy)

    @classmethod
    def _get_distributed_init_connections(
        cls,
        compiled_operation: V1CompiledOperation,
    ) -> Set[str]:
        init_connection_names = set()

        def _get_resolve_connections(replica):
            if replica and replica.init:
                return init_connection_names | set(
                    [i.connection for i in replica.init if i.connection]
                )
            return init_connection_names

        if compiled_operation.is_mpi_job_run:
            init_connection_names = _get_resolve_connections(
                compiled_operation.run.launcher
            )
            init_connection_names = _get_resolve_connections(
                compiled_operation.run.worker
            )
        elif compiled_operation.is_tf_job_run:
            init_connection_names = _get_resolve_connections(
                compiled_operation.run.chief
            )
            init_connection_names = _get_resolve_connections(
                compiled_operation.run.worker
            )
            init_connection_names = _get_resolve_connections(compiled_operation.run.ps)
            init_connection_names = _get_resolve_connections(
                compiled_operation.run.evaluator
            )
        elif (
            compiled_operation.is_pytorch_job_run
            or compiled_operation.is_paddle_job_run
        ):
            init_connection_names = _get_resolve_connections(
                compiled_operation.run.master
            )
            init_connection_names = _get_resolve_connections(
                compiled_operation.run.worker
            )
        elif compiled_operation.is_mx_job_run:
            init_connection_names = _get_resolve_connections(
                compiled_operation.run.scheduler
            )
            init_connection_names = _get_resolve_connections(
                compiled_operation.run.worker
            )
            init_connection_names = _get_resolve_connections(
                compiled_operation.run.server
            )
            init_connection_names = _get_resolve_connections(
                compiled_operation.run.tuner
            )
            init_connection_names = _get_resolve_connections(
                compiled_operation.run.tuner_tracker
            )
            init_connection_names = _get_resolve_connections(
                compiled_operation.run.tuner_server
            )
        elif compiled_operation.is_xgb_job_run:
            init_connection_names = _get_resolve_connections(
                compiled_operation.run.master
            )
            init_connection_names = _get_resolve_connections(
                compiled_operation.run.worker
            )

        return init_connection_names

    @classmethod
    def _get_init_connections(
        cls,
        compiled_operation: V1CompiledOperation,
    ) -> Set[str]:
        init_connection_names = set()
        if compiled_operation and not compiled_operation.has_pipeline:
            if compiled_operation.run.init:
                init_connection_names |= set(
                    [i.connection for i in compiled_operation.run.init if i.connection]
                )
        return init_connection_names

    @classmethod
    def get_init_connections(
        cls,
        config: V1CompiledOperation,
    ) -> Set[str]:
        if config.is_distributed_run:
            return cls._get_distributed_init_connections(
                compiled_operation=config,
            )
        else:
            return cls._get_init_connections(
                compiled_operation=config,
            )

    @classmethod
    def _get_distributed_connections(
        cls,
        compiled_operation: V1CompiledOperation,
    ) -> Set[str]:
        connections_names = set()

        def _get_resolve_connections(replica):
            if replica and replica.connections:
                return connections_names | set(replica.connections)
            return connections_names

        if compiled_operation.is_mpi_job_run:
            connections_names = _get_resolve_connections(
                compiled_operation.run.launcher
            )
            connections_names = _get_resolve_connections(compiled_operation.run.worker)
        elif compiled_operation.is_tf_job_run:
            connections_names = _get_resolve_connections(compiled_operation.run.chief)
            connections_names = _get_resolve_connections(compiled_operation.run.worker)
            connections_names = _get_resolve_connections(compiled_operation.run.ps)
            connections_names = _get_resolve_connections(
                compiled_operation.run.evaluator
            )
        elif (
            compiled_operation.is_pytorch_job_run
            or compiled_operation.is_paddle_job_run
        ):
            connections_names = _get_resolve_connections(compiled_operation.run.master)
            connections_names = _get_resolve_connections(compiled_operation.run.worker)
        elif compiled_operation.is_mx_job_run:
            connections_names = _get_resolve_connections(
                compiled_operation.run.scheduler
            )
            connections_names = _get_resolve_connections(compiled_operation.run.worker)
            connections_names = _get_resolve_connections(compiled_operation.run.server)
            connections_names = _get_resolve_connections(compiled_operation.run.tuner)
            connections_names = _get_resolve_connections(
                compiled_operation.run.tuner_tracker
            )
            connections_names = _get_resolve_connections(
                compiled_operation.run.tuner_server
            )
        elif compiled_operation.is_xgb_job_run:
            connections_names = _get_resolve_connections(compiled_operation.run.master)
            connections_names = _get_resolve_connections(compiled_operation.run.worker)

        return connections_names

    @classmethod
    def _get_connections(
        cls,
        compiled_operation: V1CompiledOperation,
    ) -> Set[str]:
        connections_names = set()
        if compiled_operation and compiled_operation.run.connections:
            connections_names |= set(compiled_operation.run.connections)
        return connections_names

    @classmethod
    def get_connections(
        cls,
        config: V1CompiledOperation,
    ) -> Set[str]:
        if config.is_distributed_run:
            return cls._get_distributed_connections(
                compiled_operation=config,
            )
        else:
            return cls._get_connections(
                compiled_operation=config,
            )

    @classmethod
    def _get_distributed_init_lineage_refs(
        cls,
        compiled_operation: V1CompiledOperation,
    ) -> Set[str]:
        init_model_version_names = set()

        def _get_resolve_models(replica):
            if replica and replica.init:
                return init_model_version_names | set(
                    [i.lineage_ref for i in replica.init if i.lineage_ref]
                )
            return init_model_version_names

        if compiled_operation.is_mpi_job_run:
            init_model_version_names = _get_resolve_models(
                compiled_operation.run.launcher
            )
            init_model_version_names = _get_resolve_models(
                compiled_operation.run.worker
            )
        elif compiled_operation.is_tf_job_run:
            init_model_version_names = _get_resolve_models(compiled_operation.run.chief)
            init_model_version_names = _get_resolve_models(
                compiled_operation.run.worker
            )
            init_model_version_names = _get_resolve_models(compiled_operation.run.ps)
            init_model_version_names = _get_resolve_models(
                compiled_operation.run.evaluator
            )
        elif (
            compiled_operation.is_pytorch_job_run
            or compiled_operation.is_paddle_job_run
        ):
            init_model_version_names = _get_resolve_models(
                compiled_operation.run.master
            )
            init_model_version_names = _get_resolve_models(
                compiled_operation.run.worker
            )
        elif compiled_operation.is_mx_job_run:
            init_model_version_names = _get_resolve_models(
                compiled_operation.run.scheduler
            )
            init_model_version_names = _get_resolve_models(
                compiled_operation.run.worker
            )
            init_model_version_names = _get_resolve_models(
                compiled_operation.run.server
            )
            init_model_version_names = _get_resolve_models(compiled_operation.run.tuner)
            init_model_version_names = _get_resolve_models(
                compiled_operation.run.tuner_tracker
            )
            init_model_version_names = _get_resolve_models(
                compiled_operation.run.tuner_server
            )
        elif compiled_operation.is_xgb_job_run:
            init_model_version_names = _get_resolve_models(
                compiled_operation.run.master
            )
            init_model_version_names = _get_resolve_models(
                compiled_operation.run.worker
            )

        return init_model_version_names

    @classmethod
    def _get_init_lineage_refs(
        cls,
        compiled_operation: V1CompiledOperation,
    ) -> Set[str]:
        init_model_version_names = set()
        if compiled_operation and not compiled_operation.has_pipeline:
            if compiled_operation.run.init:
                init_model_version_names |= set(
                    [
                        i.lineage_ref
                        for i in compiled_operation.run.init
                        if i.lineage_ref
                    ]
                )
        return init_model_version_names

    @classmethod
    def get_init_lineage_refs(
        cls,
        config: V1CompiledOperation,
    ) -> Set[str]:
        if config.is_distributed_run:
            return cls._get_distributed_init_lineage_refs(
                compiled_operation=config,
            )
        else:
            return cls._get_init_lineage_refs(
                compiled_operation=config,
            )

    @classmethod
    def _get_distributed_init_model_refs(
        cls,
        compiled_operation: V1CompiledOperation,
    ) -> Set[str]:
        init_model_version_names = set()

        def _get_resolve_models(replica):
            if replica and replica.init:
                return init_model_version_names | set(
                    [i.model_ref for i in replica.init if i.model_ref]
                )
            return init_model_version_names

        if compiled_operation.is_mpi_job_run:
            init_model_version_names = _get_resolve_models(
                compiled_operation.run.launcher
            )
            init_model_version_names = _get_resolve_models(
                compiled_operation.run.worker
            )
        elif compiled_operation.is_tf_job_run:
            init_model_version_names = _get_resolve_models(compiled_operation.run.chief)
            init_model_version_names = _get_resolve_models(
                compiled_operation.run.worker
            )
            init_model_version_names = _get_resolve_models(compiled_operation.run.ps)
            init_model_version_names = _get_resolve_models(
                compiled_operation.run.evaluator
            )
        elif (
            compiled_operation.is_pytorch_job_run
            or compiled_operation.is_paddle_job_run
        ):
            init_model_version_names = _get_resolve_models(
                compiled_operation.run.master
            )
            init_model_version_names = _get_resolve_models(
                compiled_operation.run.worker
            )
        elif compiled_operation.is_mx_job_run:
            init_model_version_names = _get_resolve_models(
                compiled_operation.run.scheduler
            )
            init_model_version_names = _get_resolve_models(
                compiled_operation.run.worker
            )
            init_model_version_names = _get_resolve_models(
                compiled_operation.run.server
            )
            init_model_version_names = _get_resolve_models(compiled_operation.run.tuner)
            init_model_version_names = _get_resolve_models(
                compiled_operation.run.tuner_tracker
            )
            init_model_version_names = _get_resolve_models(
                compiled_operation.run.tuner_server
            )
        elif compiled_operation.is_xgb_job_run:
            init_model_version_names = _get_resolve_models(
                compiled_operation.run.master
            )
            init_model_version_names = _get_resolve_models(
                compiled_operation.run.worker
            )

        return init_model_version_names

    @classmethod
    def _get_init_model_refs(
        cls,
        compiled_operation: V1CompiledOperation,
    ) -> Set[str]:
        init_model_version_names = set()
        if compiled_operation and not compiled_operation.has_pipeline:
            if compiled_operation.run.init:
                init_model_version_names |= set(
                    [i.model_ref for i in compiled_operation.run.init if i.model_ref]
                )
        return init_model_version_names

    @classmethod
    def get_init_model_refs(
        cls,
        config: V1CompiledOperation,
    ) -> Set[str]:
        if config.is_distributed_run:
            return cls._get_distributed_init_model_refs(
                compiled_operation=config,
            )
        else:
            return cls._get_init_model_refs(
                compiled_operation=config,
            )

    @classmethod
    def _get_distributed_init_artifact_refs(
        cls,
        compiled_operation: V1CompiledOperation,
    ) -> Set[str]:
        init_artifact_version_names = set()

        def _get_resolve_artifacts(replica):
            if replica and replica.init:
                return init_artifact_version_names | set(
                    [i.artifact_ref for i in replica.init if i.artifact_ref]
                )
            return init_artifact_version_names

        if compiled_operation.is_mpi_job_run:
            init_artifact_version_names = _get_resolve_artifacts(
                compiled_operation.run.launcher
            )
            init_artifact_version_names = _get_resolve_artifacts(
                compiled_operation.run.worker
            )
        elif compiled_operation.is_tf_job_run:
            init_artifact_version_names = _get_resolve_artifacts(
                compiled_operation.run.chief
            )
            init_artifact_version_names = _get_resolve_artifacts(
                compiled_operation.run.worker
            )
            init_artifact_version_names = _get_resolve_artifacts(
                compiled_operation.run.ps
            )
            init_artifact_version_names = _get_resolve_artifacts(
                compiled_operation.run.evaluator
            )
        elif (
            compiled_operation.is_pytorch_job_run
            or compiled_operation.is_paddle_job_run
        ):
            init_artifact_version_names = _get_resolve_artifacts(
                compiled_operation.run.master
            )
            init_artifact_version_names = _get_resolve_artifacts(
                compiled_operation.run.worker
            )
        elif compiled_operation.is_mx_job_run:
            init_artifact_version_names = _get_resolve_artifacts(
                compiled_operation.run.scheduler
            )
            init_artifact_version_names = _get_resolve_artifacts(
                compiled_operation.run.worker
            )
            init_artifact_version_names = _get_resolve_artifacts(
                compiled_operation.run.server
            )
            init_artifact_version_names = _get_resolve_artifacts(
                compiled_operation.run.tuner
            )
            init_artifact_version_names = _get_resolve_artifacts(
                compiled_operation.run.tuner_tracker
            )
            init_artifact_version_names = _get_resolve_artifacts(
                compiled_operation.run.tuner_server
            )
        elif compiled_operation.is_xgb_job_run:
            init_artifact_version_names = _get_resolve_artifacts(
                compiled_operation.run.master
            )
            init_artifact_version_names = _get_resolve_artifacts(
                compiled_operation.run.worker
            )

        return init_artifact_version_names

    @classmethod
    def _get_init_artifact_refs(
        cls,
        compiled_operation: V1CompiledOperation,
    ) -> Set[str]:
        init_artifact_version_names = set()
        if compiled_operation and not compiled_operation.has_pipeline:
            if compiled_operation.run.init:
                init_artifact_version_names |= set(
                    [
                        i.artifact_ref
                        for i in compiled_operation.run.init
                        if i.artifact_ref
                    ]
                )
        return init_artifact_version_names

    @classmethod
    def get_init_artifact_refs(
        cls,
        config: V1CompiledOperation,
    ) -> Set[str]:
        if config.is_distributed_run:
            return cls._get_distributed_init_artifact_refs(
                compiled_operation=config,
            )
        else:
            return cls._get_init_artifact_refs(
                compiled_operation=config,
            )

    @classmethod
    def _clean_distributed_init_version_refs(
        cls,
        compiled_operation: V1CompiledOperation,
    ) -> V1CompiledOperation:
        def _clean_resolve_refs(replica):
            if replica and replica.init:
                replica.init = [
                    i
                    for i in replica.init
                    if (
                        i.model_ref is None
                        and i.artifact_ref is None
                        and i.lineage_ref is None
                    )
                ]
                return replica
            return replica

        if compiled_operation.is_mpi_job_run:
            compiled_operation.run.launcher = _clean_resolve_refs(
                compiled_operation.run.launcher
            )
            compiled_operation.run.worker = _clean_resolve_refs(
                compiled_operation.run.worker
            )
        elif compiled_operation.is_tf_job_run:
            compiled_operation.run.chief = _clean_resolve_refs(
                compiled_operation.run.chief
            )
            compiled_operation.run.worker = _clean_resolve_refs(
                compiled_operation.run.worker
            )
            compiled_operation.run.ps = _clean_resolve_refs(compiled_operation.run.ps)
            compiled_operation.run.evaluator = _clean_resolve_refs(
                compiled_operation.run.evaluator
            )
        elif (
            compiled_operation.is_pytorch_job_run
            or compiled_operation.is_paddle_job_run
        ):
            compiled_operation.run.master = _clean_resolve_refs(
                compiled_operation.run.master
            )
            compiled_operation.run.worker = _clean_resolve_refs(
                compiled_operation.run.worker
            )
        elif compiled_operation.is_mx_job_run:
            compiled_operation.run.scheduler = _clean_resolve_refs(
                compiled_operation.run.scheduler
            )
            compiled_operation.run.worker = _clean_resolve_refs(
                compiled_operation.run.worker
            )
            compiled_operation.run.server = _clean_resolve_refs(
                compiled_operation.run.server
            )
            compiled_operation.run.tuner = _clean_resolve_refs(
                compiled_operation.run.tuner
            )
            compiled_operation.run.tuner_tracker = _clean_resolve_refs(
                compiled_operation.run.tuner_tracker
            )
            compiled_operation.run.tuner_server = _clean_resolve_refs(
                compiled_operation.run.tuner_server
            )
        elif compiled_operation.is_xgb_job_run:
            compiled_operation.run.master = _clean_resolve_refs(
                compiled_operation.run.master
            )
            compiled_operation.run.worker = _clean_resolve_refs(
                compiled_operation.run.worker
            )

        return compiled_operation

    @classmethod
    def _clean_init_version_refs(
        cls,
        compiled_operation: V1CompiledOperation,
    ) -> V1CompiledOperation:
        if compiled_operation and not compiled_operation.has_pipeline:
            if compiled_operation.run.init:
                compiled_operation.run.init = [
                    i
                    for i in compiled_operation.run.init
                    if (
                        i.model_ref is None
                        and i.artifact_ref is None
                        and i.lineage_ref is None
                    )
                ]
        return compiled_operation

    @classmethod
    def clean_init_version_refs(
        cls,
        config: V1CompiledOperation,
    ) -> V1CompiledOperation:
        if config.is_distributed_run:
            return cls._clean_distributed_init_version_refs(
                compiled_operation=config,
            )
        else:
            return cls._clean_init_version_refs(
                compiled_operation=config,
            )



from typing_extensions import Literal

from clipped.compact.pydantic import StrictStr

from polyaxon._flow.references.mixin import RefMixin
from polyaxon._schemas.base import BaseSchemaModel


class V1PathRef(BaseSchemaModel, RefMixin):
    _IDENTIFIER = "path_ref"
    _USE_DISCRIMINATOR = True

    kind: Literal[_IDENTIFIER] = _IDENTIFIER
    path: StrictStr

    def get_kind_value(self):
        return self.path



from typing import Optional

from clipped.compact.pydantic import StrictStr
from clipped.config.schema import BaseAllowSchemaModel
from clipped.types.email import EmailStr


class V1UserSingup(BaseAllowSchemaModel):
    username: Optional[StrictStr]
    email: Optional[EmailStr]
    organization: Optional[StrictStr]
    password: Optional[StrictStr]
    invitation_key: Optional[StrictStr]



import click

from clipped.formatting import Printer

from polyaxon._cli.options import (
    OPTIONS_ARTIFACT_VERSION,
    OPTIONS_NAME,
    OPTIONS_PROJECT,
)
from polyaxon._cli.project_versions import (
    copy_project_version,
    delete_project_version,
    get_project_version,
    get_project_version_stages,
    list_project_versions,
    open_project_version_dashboard,
    pull_one_or_many_project_versions,
    register_project_version,
    stage_project_version,
    transfer_project_version,
    update_project_version,
)
from polyaxon._env_vars.getters import get_project_or_local
from polyaxon._schemas.lifecycle import V1ProjectVersionKind, V1Stages
from polyaxon.logger import clean_outputs


@click.group()
@click.option(*OPTIONS_PROJECT["args"], **OPTIONS_PROJECT["kwargs"])
@click.option(*OPTIONS_ARTIFACT_VERSION["args"], **OPTIONS_ARTIFACT_VERSION["kwargs"])
@click.pass_context
@clean_outputs
def artifacts(ctx, project, version):
    """Commands for managing artifact versions."""
    ctx.obj = ctx.obj or {}
    if project or version:
        Printer.warning(
            "Passing arguments to command groups is deprecated and will be removed in v2! "
            "Please use arguments on the sub-command directly: `polyaxon ops SUB_COMMAND --help`"
        )
    ctx.obj["project"] = project
    if ctx.invoked_subcommand not in ["ls"]:
        ctx.obj["version"] = version


@artifacts.command()
@click.option(*OPTIONS_PROJECT["args"], **OPTIONS_PROJECT["kwargs"])
@click.option(
    "--query",
    "-q",
    type=str,
    help="To filter the artifact versions based on this query spec.",
)
@click.option(
    "--sort",
    "-s",
    type=str,
    help="To order the artifact versions based on the sort spec.",
)
@click.option("--limit", type=int, help="To limit the list of artifact versions.")
@click.option("--offset", type=int, help="To offset the list of artifact versions.")
@click.pass_context
@clean_outputs
def ls(ctx, project, query, sort, limit, offset):
    """List artifact versions by owner or owner/artifact.

    Examples:

    \b
    $ polyaxon artifacts ls -p project

    \b
    $ polyaxon artifacts ls --project=acme/data-versioning
    """
    owner, project_name = get_project_or_local(
        project or ctx.obj.get("project"), is_cli=True
    )
    list_project_versions(
        owner=owner,
        project_name=project_name,
        kind=V1ProjectVersionKind.ARTIFACT,
        query=query,
        sort=sort,
        limit=limit,
        offset=offset,
    )


@artifacts.command()
@click.option(*OPTIONS_PROJECT["args"], **OPTIONS_PROJECT["kwargs"])
@click.option(*OPTIONS_ARTIFACT_VERSION["args"], **OPTIONS_ARTIFACT_VERSION["kwargs"])
@click.option("--description", type=str, help="Description of the version.")
@click.option("--tags", type=str, help="Tags of the version (comma separated values).")
@click.option(
    "--content",
    type=str,
    help="Additional content/metadata (Json object) to save with the artifact version.",
)
@click.option("--run-uid", type=str, help="The run to promote as a artifact version.")
@click.option(
    "--artifacts",
    "artifacts_",
    type=str,
    help="The artifacts to link to this artifact version.",
)
@click.option(
    "--connection", type=str, help="The connection to link to this artifact version."
)
@click.option(
    "--force",
    is_flag=True,
    default=False,
    help="Flag to force register if the version already exists.",
)
@click.pass_context
@clean_outputs
def register(
    ctx,
    project,
    version,
    description,
    tags,
    content,
    run_uid,
    artifacts_,
    connection,
    force,
):
    """Register a new artifact version.
    If the name corresponds to an existing artifact version,
    it will raise an error or it will update the version if `--force` is provided.

    Examples:

    \b
    $ polyaxon artifacts register --version=version-name --connection=dataset-gcs --run-uid=uuid

    \b
    $ polyaxon artifacts register --project=images-dataset --connection=dataset-gcs --description="..."

    \b
    $ polyaxon artifacts register -p images-dataset --content='{"foo": "bar"}' -ver latest --run-uid=uuid --artifacts=lin1,lin2

    \b
    $ polyaxon artifacts register -p owner/name -ver v1 --tags="tag1,tag2"
    """
    version = version or ctx.obj.get("version")
    owner, project_name = get_project_or_local(
        project or ctx.obj.get("project"), is_cli=True
    )
    register_project_version(
        owner=owner,
        project_name=project_name,
        version=version,
        kind=V1ProjectVersionKind.ARTIFACT,
        description=description,
        tags=tags,
        content=content,
        run=run_uid,
        connection=connection,
        artifacts=artifacts_,
        force=force,
    )


@artifacts.command()
@click.option(*OPTIONS_PROJECT["args"], **OPTIONS_PROJECT["kwargs"])
@click.option(*OPTIONS_ARTIFACT_VERSION["args"], **OPTIONS_ARTIFACT_VERSION["kwargs"])
@click.option(
    "--to-project",
    "-to",
    help="The project to transfer the version to.",
)
@click.option(
    *OPTIONS_NAME["args"],
    type=str,
    help="Name of the artifact version, must be unique within the same project.",
)
@click.option(
    "--description", type=str, help="Optional new description of the version."
)
@click.option(
    "--tags",
    type=str,
    help="Optional new tags of the version (comma separated values).",
)
@click.option(
    "--content",
    type=str,
    help="Optional new content/metadata (Json object) to save with the artifact version.",
)
@click.option(
    "--force",
    is_flag=True,
    default=False,
    help="Flag to force copy if the version already exists.",
)
@click.pass_context
@clean_outputs
def copy(
    ctx,
    project,
    version,
    to_project,
    name,
    description,
    tags,
    content,
    force,
):
    """Copy an artifact version.
    If the name corresponds to an existing artifact version,
    it will raise an error or it will update the version if `--force` is provided.

    Examples:

    \b
    $ polyaxon artifacts copy --version=version-name --to-project dest-project

    \b
    $ polyaxon artifacts copy --project=images-dataset -to dest-project --force

    \b
    $ polyaxon artifacts copy -p images-dataset --content='{"foo": "bar"}' -ver latest

    \b
    $ polyaxon artifacts copy -p owner/name -ver v1 --tags="tag1,tag2" --name new-v1
    """
    version = version or ctx.obj.get("version")
    owner, project_name = get_project_or_local(
        project or ctx.obj.get("project"), is_cli=True
    )
    copy_project_version(
        owner=owner,
        project_name=project_name,
        version=version,
        kind=V1ProjectVersionKind.ARTIFACT,
        to_project=to_project,
        name=name,
        description=description,
        tags=tags,
        content=content,
        force=force,
    )


@artifacts.command()
@click.option(*OPTIONS_PROJECT["args"], **OPTIONS_PROJECT["kwargs"])
@click.option(*OPTIONS_ARTIFACT_VERSION["args"], **OPTIONS_ARTIFACT_VERSION["kwargs"])
@click.pass_context
@clean_outputs
def get(ctx, project, version):
    """Get info for an artifact version by name.

    Examples:

    To get a default artifact version:

    \b
    $ polyaxon artifacts get -p data-versioning

    To get by specific owner/name

    \b
    $ polyaxon artifacts get -p owner/data-versioning -ver rc1
    """
    version = version or ctx.obj.get("version") or "latest"
    owner, project_name = get_project_or_local(
        project or ctx.obj.get("project"), is_cli=True
    )
    get_project_version(
        owner=owner,
        project_name=project_name,
        kind=V1ProjectVersionKind.ARTIFACT,
        version=version,
    )


@artifacts.command()
@click.option(*OPTIONS_PROJECT["args"], **OPTIONS_PROJECT["kwargs"])
@click.option(*OPTIONS_ARTIFACT_VERSION["args"], **OPTIONS_ARTIFACT_VERSION["kwargs"])
@click.pass_context
@clean_outputs
def stages(ctx, project, version):
    """List stages information for an artifact version by name, name & version, owner/name & tag.

    Examples:

    \b
    $ polyaxon artifacts stages // returns `latest` in current project

    \b
    $ polyaxon artifacts stages --project=my-project --version=test-version

    \b
    $ polyaxon artifacts stages -p owner/my-project -ver rc12
    """
    version = version or ctx.obj.get("version") or "latest"
    owner, project_name = get_project_or_local(
        project or ctx.obj.get("project"), is_cli=True
    )
    get_project_version_stages(
        owner=owner,
        project_name=project_name,
        kind=V1ProjectVersionKind.ARTIFACT,
        version=version,
    )


@artifacts.command()
@click.option(*OPTIONS_PROJECT["args"], **OPTIONS_PROJECT["kwargs"])
@click.option(*OPTIONS_ARTIFACT_VERSION["args"], **OPTIONS_ARTIFACT_VERSION["kwargs"])
@click.option(
    "--yes",
    "-y",
    is_flag=True,
    default=False,
    help="Automatic yes to prompts. "
    'Assume "yes" as answer to all prompts and run non-interactively.',
)
@click.pass_context
@clean_outputs
def delete(ctx, project, version, yes):
    """Delete a artifact version.
    \b
    $ polyaxon artifacts delete  // delete `latest` in current project

    \b
    $ polyaxon artifacts delete --project=my-project --version=test-version

    \b
    $ polyaxon artifacts get -p owner/my-project -ver rc12
    """
    version = version or ctx.obj.get("version") or "latest"
    owner, project_name = get_project_or_local(
        project or ctx.obj.get("project"), is_cli=True
    )
    delete_project_version(
        owner=owner,
        project_name=project_name,
        kind=V1ProjectVersionKind.ARTIFACT,
        version=version,
        yes=yes,
    )


@artifacts.command()
@click.option(*OPTIONS_PROJECT["args"], **OPTIONS_PROJECT["kwargs"])
@click.option(*OPTIONS_ARTIFACT_VERSION["args"], **OPTIONS_ARTIFACT_VERSION["kwargs"])
@click.option(
    *OPTIONS_NAME["args"],
    type=str,
    help="Name of the artifact version, must be unique within the same project.",
)
@click.option("--description", type=str, help="Description of the artifact version.")
@click.option(
    "--tags", type=str, help="Tags of the artifact version (comma separated values)."
)
@click.pass_context
@clean_outputs
def update(ctx, project, version, name, description, tags):
    """Update artifact version.

    Uses /docs/core/cli/#caching

    Examples:

    \b
    $ polyaxon artifacts update --version=foobar --description="..."

    \b
    $ polyaxon artifacts update -p mike1/foobar -ver current-name --name=new-name

    \b
    $ polyaxon artifacts update --tags="foo, bar"
    """
    version = version or ctx.obj.get("version") or "latest"
    owner, project_name = get_project_or_local(
        project or ctx.obj.get("project"), is_cli=True
    )
    update_project_version(
        owner=owner,
        project_name=project_name,
        kind=V1ProjectVersionKind.ARTIFACT,
        version=version,
        name=name,
        description=description,
        tags=tags,
    )


@artifacts.command()
@click.option(*OPTIONS_PROJECT["args"], **OPTIONS_PROJECT["kwargs"])
@click.option(*OPTIONS_ARTIFACT_VERSION["args"], **OPTIONS_ARTIFACT_VERSION["kwargs"])
@click.option(
    "--to",
    "-to",
    type=click.Choice(V1Stages.to_list(), case_sensitive=True),
    help="Stage to transition to.",
)
@click.option(
    "--reason",
    type=str,
    help="Reason or service to set with this stage change. "
    "Default: UserStageUpdate, e.g. CustomAlertTrigger.",
)
@click.option(
    "--message", type=str, help="Additional information to set with this stage change."
)
@click.pass_context
@clean_outputs
def stage(ctx, project, version, to, reason, message):
    """Update stage for a artifact version.

    Uses /docs/core/cli/#caching

    Examples:

    \b
    $ polyaxon artifacts stage -ver rc12 -to production

    \b
    $ polyaxon artifacts stage -p acme/foobar -ver rc12 --to=staging --reason GithubAction --message="Use carefully!"
    """
    version = version or ctx.obj.get("version") or "latest"
    owner, project_name = get_project_or_local(
        project or ctx.obj.get("project"), is_cli=True
    )
    stage_project_version(
        owner=owner,
        project_name=project_name,
        kind=V1ProjectVersionKind.ARTIFACT,
        version=version,
        to=to,
        reason=reason,
        message=message,
    )


@artifacts.command()
@click.option(*OPTIONS_PROJECT["args"], **OPTIONS_PROJECT["kwargs"])
@click.option(*OPTIONS_ARTIFACT_VERSION["args"], **OPTIONS_ARTIFACT_VERSION["kwargs"])
@click.option(
    "--to-project",
    "-to",
    help="The project to transfer the version to.",
)
@click.pass_context
@clean_outputs
def transfer(ctx, project, version, to_project):
    """Transfer the version to a destination project under the same owner/organization.

    Uses /docs/core/cli/#caching

    Examples:

    \b
    $ polyaxon artifacts transfer -ver rc12 -to dest-project

    \b
    $ polyaxon artifacts transfer -p acme/foobar -ver rc12 --to-project=dest-project
    """
    version = version or ctx.obj.get("version") or "latest"
    owner, project_name = get_project_or_local(
        project or ctx.obj.get("project"), is_cli=True
    )
    transfer_project_version(
        owner=owner,
        project_name=project_name,
        kind=V1ProjectVersionKind.ARTIFACT,
        version=version,
        to_project=to_project,
    )


@artifacts.command()
@click.option(*OPTIONS_PROJECT["args"], **OPTIONS_PROJECT["kwargs"])
@click.option(*OPTIONS_ARTIFACT_VERSION["args"], **OPTIONS_ARTIFACT_VERSION["kwargs"])
@click.option(
    "--all-versions",
    "-a",
    is_flag=True,
    default=False,
    help="To pull all versions.",
)
@click.option(
    "--query", "-q", type=str, help="To filter the versions based on a query spec."
)
@click.option("--limit", "-l", type=int, help="To limit the list of runs.")
@click.option("--offset", "-off", type=int, help="To offset the list of runs.")
@click.option(
    "--path",
    "--path-to",
    type=click.Path(exists=False),
    help="Optional path where the artifact versions are persisted, "
    "default value is taken from the env var: `POLYAXON_OFFLINE_ROOT`.",
)
@click.option(
    "--no-artifacts",
    is_flag=True,
    default=False,
    help="To disable downloading the run's artifacts and only persist the metadata.",
)
@click.pass_context
@clean_outputs
def pull(
    ctx,
    project,
    version,
    all_versions,
    query,
    limit,
    offset,
    path,
    no_artifacts,
):
    """Package and download an artifact version or many artifact versions
    to a local path with or without assets.

    Uses /docs/core/cli/#caching

    Examples:

    \b
    $ polyaxon artifacts pull -ver rc12

    \b
    $ polyaxon artifacts pull -p acme/foobar -q "stage: production, name: %-v1%"

    \b
    $ polyaxon artifacts pull -p acme/foobar -a --path /tmp/versions
    """
    owner, project_name = get_project_or_local(
        project or ctx.obj.get("project"), is_cli=True
    )
    pull_one_or_many_project_versions(
        owner=owner,
        project_name=project_name,
        kind=V1ProjectVersionKind.ARTIFACT,
        version=version,
        all_versions=all_versions,
        query=query,
        limit=limit,
        offset=offset,
        path=path,
        download_artifacts=not no_artifacts,
    )


@artifacts.command()
@click.option(*OPTIONS_PROJECT["args"], **OPTIONS_PROJECT["kwargs"])
@click.option(*OPTIONS_ARTIFACT_VERSION["args"], **OPTIONS_ARTIFACT_VERSION["kwargs"])
@click.option(
    "--yes",
    "-y",
    is_flag=True,
    default=False,
    help="Automatic yes to prompts. "
    'Assume "yes" as answer to all prompts and run non-interactively.',
)
@click.option(
    "--url",
    is_flag=True,
    default=False,
    help="Print the url of the dashboard for this artifact version.",
)
@click.pass_context
@clean_outputs
def dashboard(ctx, project, version, yes, url):
    """Open this artifact version's dashboard details in browser."""
    version = version or ctx.obj.get("version") or "latest"
    owner, project_name = get_project_or_local(
        project or ctx.obj.get("project"), is_cli=True
    )
    open_project_version_dashboard(
        owner=owner,
        project_name=project_name,
        kind=V1ProjectVersionKind.ARTIFACT,
        version=version,
        url=url,
        yes=yes,
    )



from typing import Dict, Union

from polyaxon._auxiliaries import get_default_notification_container
from polyaxon._flow import (
    V1IO,
    V1Component,
    V1NotifierJob,
    V1Operation,
    V1Param,
    V1Plugins,
    V1Termination,
)


def get_notifier_operation(
    connection: str,
    backend: str,
    owner: str,
    project: str,
    run_uuid: str,
    run_name: str,
    condition: Union[str, Dict],
) -> V1Operation:
    return V1Operation(
        params={
            "backend": V1Param.construct(value=backend),
            "owner": V1Param.construct(value=owner),
            "project": V1Param.construct(value=project),
            "uuid": V1Param.construct(value=run_uuid),
            "name": V1Param.construct(value=run_name),
            "condition": V1Param.construct(value=condition),
        },
        termination=V1Termination.construct(max_retries=3),
        component=V1Component.construct(
            name="notifier",
            plugins=V1Plugins.construct(
                auth=False,
                collect_logs=False,
                collect_artifacts=False,
                collect_resources=False,
                auto_resume=False,
                sync_statuses=False,
                external_host=True,
            ),
            inputs=[
                V1IO.construct(name="backend", type="str", is_optional=False),
                V1IO.construct(name="owner", type="str", is_optional=False),
                V1IO.construct(name="project", type="str", is_optional=False),
                V1IO.construct(name="uuid", type="str", is_optional=False),
                V1IO.construct(name="name", type="str", is_optional=True),
                V1IO.construct(name="condition", type="dict", is_optional=True),
                V1IO.construct(name="connection", type="str", is_optional=True),
            ],
            run=V1NotifierJob.construct(
                connections=[connection],
                container=get_default_notification_container(),
            ),
        ),
    )



from polyaxon._sdk.schemas.v1_activity import V1Activity
from polyaxon._sdk.schemas.v1_agent import V1Agent
from polyaxon._sdk.schemas.v1_agent_state_response import V1AgentStateResponse
from polyaxon._sdk.schemas.v1_agent_state_response_agent_state import (
    V1AgentStateResponseAgentState,
)
from polyaxon._sdk.schemas.v1_agent_status_body_request import V1AgentStatusBodyRequest
from polyaxon._sdk.schemas.v1_analytics_spec import V1AnalyticsSpec
from polyaxon._sdk.schemas.v1_artifact_tree import V1ArtifactTree
from polyaxon._sdk.schemas.v1_auth import V1Auth
from polyaxon._sdk.schemas.v1_cloning import V1Cloning
from polyaxon._sdk.schemas.v1_connection_response import V1ConnectionResponse
from polyaxon._sdk.schemas.v1_dashboard import V1Dashboard
from polyaxon._sdk.schemas.v1_dashboard_spec import V1DashboardSpec
from polyaxon._sdk.schemas.v1_entities_tags import V1EntitiesTags
from polyaxon._sdk.schemas.v1_entities_transfer import V1EntitiesTransfer
from polyaxon._sdk.schemas.v1_entity_notification_body import V1EntityNotificationBody
from polyaxon._sdk.schemas.v1_entity_stage_body_request import V1EntityStageBodyRequest
from polyaxon._sdk.schemas.v1_entity_status_body_request import (
    V1EntityStatusBodyRequest,
)
from polyaxon._sdk.schemas.v1_events_response import V1EventsResponse
from polyaxon._sdk.schemas.v1_list_activities_response import V1ListActivitiesResponse
from polyaxon._sdk.schemas.v1_list_agents_response import V1ListAgentsResponse
from polyaxon._sdk.schemas.v1_list_bookmarks_response import V1ListBookmarksResponse
from polyaxon._sdk.schemas.v1_list_connections_response import V1ListConnectionsResponse
from polyaxon._sdk.schemas.v1_list_dashboards_response import V1ListDashboardsResponse
from polyaxon._sdk.schemas.v1_list_organization_members_response import (
    V1ListOrganizationMembersResponse,
)
from polyaxon._sdk.schemas.v1_list_organizations_response import (
    V1ListOrganizationsResponse,
)
from polyaxon._sdk.schemas.v1_list_presets_response import V1ListPresetsResponse
from polyaxon._sdk.schemas.v1_list_project_versions_response import (
    V1ListProjectVersionsResponse,
)
from polyaxon._sdk.schemas.v1_list_projects_response import V1ListProjectsResponse
from polyaxon._sdk.schemas.v1_list_queues_response import V1ListQueuesResponse
from polyaxon._sdk.schemas.v1_list_run_artifacts_response import (
    V1ListRunArtifactsResponse,
)
from polyaxon._sdk.schemas.v1_list_run_connections_response import (
    V1ListRunConnectionsResponse,
)
from polyaxon._sdk.schemas.v1_list_run_edges_response import V1ListRunEdgesResponse
from polyaxon._sdk.schemas.v1_list_runs_response import V1ListRunsResponse
from polyaxon._sdk.schemas.v1_list_searches_response import V1ListSearchesResponse
from polyaxon._sdk.schemas.v1_list_service_accounts_response import (
    V1ListServiceAccountsResponse,
)
from polyaxon._sdk.schemas.v1_list_tags_response import V1ListTagsResponse
from polyaxon._sdk.schemas.v1_list_team_members_response import (
    V1ListTeamMembersResponse,
)
from polyaxon._sdk.schemas.v1_list_teams_response import V1ListTeamsResponse
from polyaxon._sdk.schemas.v1_list_token_response import V1ListTokenResponse
from polyaxon._sdk.schemas.v1_operation_body import V1OperationBody
from polyaxon._sdk.schemas.v1_organization import V1Organization
from polyaxon._sdk.schemas.v1_organization_member import V1OrganizationMember
from polyaxon._sdk.schemas.v1_password_change import V1PasswordChange
from polyaxon._sdk.schemas.v1_pipeline import V1Pipeline
from polyaxon._sdk.schemas.v1_preset import V1Preset
from polyaxon._sdk.schemas.v1_project import V1Project
from polyaxon._sdk.schemas.v1_project_settings import V1ProjectSettings
from polyaxon._sdk.schemas.v1_project_user_access import V1ProjectUserAccess
from polyaxon._sdk.schemas.v1_project_version import V1ProjectVersion
from polyaxon._sdk.schemas.v1_queue import V1Queue
from polyaxon._sdk.schemas.v1_run import V1Run
from polyaxon._sdk.schemas.v1_run_connection import V1RunConnection
from polyaxon._sdk.schemas.v1_run_edge import V1RunEdge
from polyaxon._sdk.schemas.v1_run_edge_lineage import V1RunEdgeLineage
from polyaxon._sdk.schemas.v1_run_edges_graph import V1RunEdgesGraph
from polyaxon._sdk.schemas.v1_run_reference_catalog import V1RunReferenceCatalog
from polyaxon._sdk.schemas.v1_run_settings import V1RunSettings
from polyaxon._sdk.schemas.v1_search import V1Search
from polyaxon._sdk.schemas.v1_search_spec import V1SearchSpec
from polyaxon._sdk.schemas.v1_section_spec import V1SectionSpec
from polyaxon._sdk.schemas.v1_service_account import V1ServiceAccount
from polyaxon._sdk.schemas.v1_settings_catalog import V1SettingsCatalog
from polyaxon._sdk.schemas.v1_tag import V1Tag
from polyaxon._sdk.schemas.v1_team import V1Team
from polyaxon._sdk.schemas.v1_team_member import V1TeamMember
from polyaxon._sdk.schemas.v1_team_settings import V1TeamSettings
from polyaxon._sdk.schemas.v1_token import V1Token
from polyaxon._sdk.schemas.v1_trial_start import V1TrialStart
from polyaxon._sdk.schemas.v1_user import V1User
from polyaxon._sdk.schemas.v1_user_email import V1UserEmail
from polyaxon._sdk.schemas.v1_user_singup import V1UserSingup
from polyaxon._sdk.schemas.v1_uuids import V1Uuids



from polyaxon._flow.references.dag import V1DagRef
from polyaxon._flow.references.hub import V1HubRef
from polyaxon._flow.references.mixin import RefMixin
from polyaxon._flow.references.path import V1PathRef
from polyaxon._flow.references.url import V1UrlRef



NAME = "polyaxon"
VERSION = "2.0.0-rc54"
SCHEMA_VERSION = 1.1
DESC = "Command Line Interface (CLI) and client to interact with Polyaxon API."
URL = "https://github.com/polyaxon/polyaxon"
AUTHOR = "Polyaxon, Inc."
EMAIL = "contact@polyaxon.com"
LICENSE = "Apache 2.0"



from typing import Dict, Optional, Union

import urllib3

from clipped.compact.pydantic import Extra, Field, StrictStr
from clipped.utils.http import clean_host

from polyaxon._contexts import paths as ctx_paths
from polyaxon._env_vars.keys import (
    ENV_KEYS_API_VERSION,
    ENV_KEYS_ARCHIVES_ROOT,
    ENV_KEYS_ASSERT_HOSTNAME,
    ENV_KEYS_AUTHENTICATION_TYPE,
    ENV_KEYS_CERT_FILE,
    ENV_KEYS_CONNECTION_POOL_MAXSIZE,
    ENV_KEYS_DEBUG,
    ENV_KEYS_DISABLE_ERRORS_REPORTING,
    ENV_KEYS_HEADER,
    ENV_KEYS_HEADER_SERVICE,
    ENV_KEYS_HOST,
    ENV_KEYS_INTERVAL,
    ENV_KEYS_INTERVALS_COMPATIBILITY_CHECK,
    ENV_KEYS_IS_MANAGED,
    ENV_KEYS_IS_OFFLINE,
    ENV_KEYS_K8S_IN_CLUSTER,
    ENV_KEYS_K8S_NAMESPACE,
    ENV_KEYS_KEY_FILE,
    ENV_KEYS_LOG_LEVEL,
    ENV_KEYS_NO_API,
    ENV_KEYS_NO_OP,
    ENV_KEYS_RETRIES,
    ENV_KEYS_SSL_CA_CERT,
    ENV_KEYS_TIME_ZONE,
    ENV_KEYS_TIMEOUT,
    ENV_KEYS_TRACKING_TIMEOUT,
    ENV_KEYS_VERIFY_SSL,
    ENV_KEYS_WATCH_INTERVAL,
)
from polyaxon._schemas.base import BaseSchemaModel
from polyaxon._sdk.configuration import Configuration
from polyaxon._services.auth import AuthenticationTypes
from polyaxon._services.headers import PolyaxonServiceHeaders
from polyaxon._services.values import PolyaxonServices
from polyaxon.api import LOCALHOST, POLYAXON_CLOUD_HOST
from polyaxon.exceptions import PolyaxonClientException
from polyaxon.pkg import VERSION


class ClientConfig(BaseSchemaModel):
    _IDENTIFIER = "global"

    _PAGE_SIZE = 20
    _BASE_URL = "{}/api/{}"

    host: Optional[StrictStr] = Field(alias=ENV_KEYS_HOST)
    version: Optional[StrictStr] = Field(default="v1", alias=ENV_KEYS_API_VERSION)
    debug: Optional[bool] = Field(default=False, alias=ENV_KEYS_DEBUG)
    log_level: Optional[StrictStr] = Field(alias=ENV_KEYS_LOG_LEVEL)
    authentication_type: Optional[StrictStr] = Field(
        default=AuthenticationTypes.TOKEN, alias=ENV_KEYS_AUTHENTICATION_TYPE
    )
    is_managed: Optional[bool] = Field(default=False, alias=ENV_KEYS_IS_MANAGED)
    is_offline: Optional[bool] = Field(default=False, alias=ENV_KEYS_IS_OFFLINE)
    in_cluster: Optional[bool] = Field(default=False, alias=ENV_KEYS_K8S_IN_CLUSTER)
    no_op: Optional[bool] = Field(default=False, alias=ENV_KEYS_NO_OP)
    timeout: Optional[float] = Field(default=20, alias=ENV_KEYS_TIMEOUT)
    tracking_timeout: Optional[float] = Field(
        default=1, alias=ENV_KEYS_TRACKING_TIMEOUT
    )
    timezone: Optional[StrictStr] = Field(alias=ENV_KEYS_TIME_ZONE)
    watch_interval: Optional[int] = Field(default=5, alias=ENV_KEYS_WATCH_INTERVAL)
    interval: Optional[float] = Field(default=5, alias=ENV_KEYS_INTERVAL)
    verify_ssl: Optional[bool] = Field(alias=ENV_KEYS_VERIFY_SSL)
    ssl_ca_cert: Optional[StrictStr] = Field(alias=ENV_KEYS_SSL_CA_CERT)
    cert_file: Optional[StrictStr] = Field(alias=ENV_KEYS_CERT_FILE)
    key_file: Optional[StrictStr] = Field(alias=ENV_KEYS_KEY_FILE)
    assert_hostname: Optional[bool] = Field(alias=ENV_KEYS_ASSERT_HOSTNAME)
    connection_pool_maxsize: Optional[int] = Field(
        alias=ENV_KEYS_CONNECTION_POOL_MAXSIZE
    )
    archives_root: Optional[StrictStr] = Field(
        default=ctx_paths.CONTEXT_ARCHIVES_ROOT, alias=ENV_KEYS_ARCHIVES_ROOT
    )
    header: Optional[Union[StrictStr, PolyaxonServiceHeaders]] = Field(
        alias=ENV_KEYS_HEADER
    )
    header_service: Optional[Union[StrictStr, PolyaxonServices]] = Field(
        alias=ENV_KEYS_HEADER_SERVICE
    )
    namespace: Optional[StrictStr] = Field(alias=ENV_KEYS_K8S_NAMESPACE)
    no_api: Optional[bool] = Field(default=False, alias=ENV_KEYS_NO_API)
    disable_errors_reporting: Optional[bool] = Field(
        default=False, alias=ENV_KEYS_DISABLE_ERRORS_REPORTING
    )
    compatibility_check_interval: Optional[int] = Field(
        alias=ENV_KEYS_INTERVALS_COMPATIBILITY_CHECK
    )
    retries: Optional[int] = Field(alias=ENV_KEYS_RETRIES)
    token: Optional[StrictStr]
    client_header: Optional[Dict]

    class Config:
        extra = Extra.ignore

    def __init__(
        self,
        host: Optional[str] = None,
        token: Optional[str] = None,
        use_cloud_host: bool = False,
        retries: Optional[int] = None,
        **data
    ):
        host = (
            clean_host(host or LOCALHOST) if not use_cloud_host else POLYAXON_CLOUD_HOST
        )
        super().__init__(host=host, **data)
        self.retries = retries if not use_cloud_host else 0
        self.token = token
        self.client_header = {}
        if all([self.header, self.header_service]):
            self.client_header["header_name"] = self.header
            self.client_header["header_value"] = self.header_service

    @property
    def base_url(self) -> str:
        return self._BASE_URL.format(clean_host(self.host), self.version)

    def set_cli_header(self) -> None:
        self.header = PolyaxonServiceHeaders.get_header(PolyaxonServiceHeaders.SERVICE)
        self.header_service = VERSION
        self.client_header["header_name"] = self.header
        self.client_header["header_value"] = self.header_service

    def set_agent_header(self) -> None:
        self.header = PolyaxonServiceHeaders.get_header(PolyaxonServiceHeaders.SERVICE)
        self.header_service = PolyaxonServices.AGENT
        self.client_header["header_name"] = self.header
        self.client_header["header_value"] = self.header_service

    def get_full_headers(self, headers=None, auth_key="Authorization") -> Dict:
        request_headers = {}
        request_headers.update(headers or {})
        request_headers.update(self.client_header or {})

        if auth_key not in request_headers and self.token:
            request_headers.update(
                {auth_key: "{} {}".format(self.authentication_type, self.token)}
            )
        if self.header and self.header_service:
            request_headers.update({self.header: self.header_service})
        return request_headers

    @property
    def sdk_config(self) -> Configuration:
        if not self.host and not self.in_cluster:
            raise PolyaxonClientException(
                "Api config requires at least a host if not running in-cluster."
            )

        config = Configuration()
        config.retries = self.retries
        config.debug = self.debug
        config.host = clean_host(self.host)
        config.verify_ssl = self.verify_ssl
        if not config.verify_ssl:
            urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)
        config.ssl_ca_cert = self.ssl_ca_cert
        config.cert_file = self.cert_file
        config.key_file = self.key_file
        config.assert_hostname = self.assert_hostname
        if self.connection_pool_maxsize:
            config.connection_pool_maxsize = self.connection_pool_maxsize
        if self.token:
            config.api_key["ApiKey"] = self.token
            config.api_key_prefix["ApiKey"] = self.authentication_type
        return config

    @property
    def async_sdk_config(self) -> Configuration:
        config = self.sdk_config
        config.connection_pool_maxsize = 100
        return config

    @classmethod
    def patch_from(cls, config: "ClientConfig", **kwargs) -> "ClientConfig":
        data = {**config.to_dict(), **kwargs}
        return cls.from_dict(data)



import os

from typing import Optional, Tuple

from polyaxon._env_vars.keys import (
    ENV_KEYS_AGENT_INSTANCE,
    ENV_KEYS_ARTIFACTS_STORE_NAME,
)
from polyaxon.exceptions import PolyaxonAgentError


def get_agent_info(agent_instance: Optional[str] = None) -> Tuple[str, str]:
    agent_instance = agent_instance or os.getenv(ENV_KEYS_AGENT_INSTANCE, None)
    if not agent_instance:
        raise PolyaxonAgentError(
            "Could get agent info, "
            "please make sure that this agent was registered in Polyaxon."
        )

    parts = agent_instance.split(".")
    if not len(parts) == 3 or parts[1] != "agents":
        raise PolyaxonAgentError(
            "agent instance is invalid `{}`, "
            "please make sure that this agent was registered in Polyaxon.".format(
                agent_instance
            )
        )
    return parts[0], parts[-1]


def get_artifacts_store_name(default: Optional[str] = "artifacts_store"):
    """Get the artifacts store name"""
    return os.getenv(ENV_KEYS_ARTIFACTS_STORE_NAME, default)



from datetime import datetime
from typing import Any, Dict, Iterable, Optional

from clipped.utils.tz import now
from kubernetes.client.rest import ApiException

from polyaxon._k8s.manager.manager import K8sManager
from polyaxon.client import RunClient
from polyaxon.exceptions import PolyaxonK8sError
from traceml.logging import V1Log


def query_logs(
    k8s_manager: K8sManager,
    pod_id: str,
    container_id: str,
    stream: bool = False,
    since_seconds: Optional[int] = None,
) -> Any:
    params: Dict[str, Any] = {}
    if stream:
        params = {"follow": True, "_preload_content": False}
    if since_seconds:
        params = {"since_seconds": since_seconds}

    return k8s_manager.k8s_api.read_namespaced_pod_log(
        pod_id, k8s_manager.namespace, container=container_id, timestamps=True, **params
    )


def process_log_line(log_line: str):
    if not isinstance(log_line, str):
        log_line = log_line.decode("utf-8")

    return V1Log.process_log_line(
        value=log_line.strip(), node=None, pod=None, container=None
    )


def stream_logs(
    k8s_manager: K8sManager, pod_id: str, container_id: str
) -> Iterable[str]:
    raw = None
    retries = 0
    no_logs = True

    while retries < 3 and no_logs:
        try:
            raw = query_logs(
                k8s_manager=k8s_manager,
                pod_id=pod_id,
                container_id=container_id,
                stream=True,
            )
        except (PolyaxonK8sError, ApiException):
            retries += 1

    if not raw:
        yield ""
    else:
        for log_line in raw.stream():
            if log_line:
                yield process_log_line(log_line=log_line)


def process_logs(
    k8s_manager: K8sManager,
    pod_id: str,
    container_id: str,
    filepath: str,
    since_seconds: Optional[int],
) -> bool:
    logs = None
    retries = 0
    no_logs = True
    while retries < 3 and no_logs:
        try:
            logs = query_logs(
                k8s_manager=k8s_manager,
                pod_id=pod_id,
                container_id=container_id,
                since_seconds=since_seconds,
            )
            no_logs = False
        except (PolyaxonK8sError, ApiException):
            retries += 1

    if not logs:
        return False

    log_lines = []
    for log_line in logs.split("\n"):
        if log_line:
            log_lines.append(process_log_line(log_line=log_line))

    # Creating the new file
    if not log_lines:
        return False

    with open(filepath, "w+") as destination:
        destination.write("\n".join(log_lines))

    return True


def sync_logs(
    k8s_manager: K8sManager,
    client: RunClient,
    last_check: Optional[datetime],
    pod_id: str,
    container_id: str,
    owner: str,
    project: str,
    run_uuid: str,
):
    new_check = now()
    since_seconds = None
    if last_check:
        since_seconds = (new_check - last_check).total_seconds()

        if since_seconds < 1:
            return last_check

    filepath = str(new_check.timestamp())
    created = process_logs(
        k8s_manager=k8s_manager,
        pod_id=pod_id,
        container_id=container_id,
        since_seconds=since_seconds,
        filepath=filepath,
    )
    if created:
        client.client.upload_run_logs(
            owner, project, run_uuid, uploadfile=filepath, path=filepath
        )
        return new_check

    return last_check



from typing import Dict, List

from polyaxon._compiler.lineage.io_collector import collect_io_artifacts
from polyaxon._connections import V1Connection
from polyaxon._flow import ParamSpec, V1CompiledOperation
from traceml.artifacts import V1RunArtifact


def resolve_artifacts_lineage(
    owner_name: str,
    project_name: str,
    project_uuid: str,
    run_uuid: str,
    run_name: str,
    run_path: str,
    param_spec: Dict[str, ParamSpec],
    compiled_operation: V1CompiledOperation,
    artifacts_store: V1Connection,
    connection_by_names: Dict[str, V1Connection],
) -> List[V1RunArtifact]:
    return collect_io_artifacts(
        compiled_operation=compiled_operation, connection_by_names=connection_by_names
    )



VOLUME_MOUNT_DOCKER = "docker"
VOLUME_MOUNT_SHM = "shm"
VOLUME_MOUNT_CONFIGS = "configs-context"
VOLUME_MOUNT_ARTIFACTS = "artifacts-context"
VOLUME_MOUNT_CONNECTIONS_FORMAT = "connections-context-{}"



import click

from clipped.formatting import Printer

from polyaxon.logger import clean_outputs


@click.command()
@click.argument("shell")
@clean_outputs
def completion(shell):
    """Show documentation for installing the auto-completion for polyaxon cli.

    Valid options: [bash, zsh, fish]

    \b
    $ polyaxon completion SHELL
    """

    if shell == "bash":
        Printer.heading(
            "Please save the following scripts:",
        )
        Printer.print(
            "_POLYAXON_COMPLETE=bash_source polyaxon > ~/.polyaxon-complete.bash"
        )
        Printer.print("_PLX_COMPLETE=bash_source plx > ~/.plx-complete.bash")
        Printer.heading(
            "Add the following lines to your: `~/.bashrc`",
        )
        Printer.print_text(
            "# Polyaxon completion\n. ~/.polyaxon-complete.bash\n. ~/.plx-complete.bash"
        )
        Printer.heading(
            "Reload your shell.",
        )
    elif shell == "zsh":
        Printer.heading(
            "Please save the following scripts:",
        )
        Printer.print(
            "_POLYAXON_COMPLETE=zsh_source polyaxon > ~/.polyaxon-complete.zsh"
        )
        Printer.print("_PLX_COMPLETE=zsh_source plx > ~/.plx-complete.zsh")
        Printer.heading(
            "Add the following lines to your: `~/.zshrc`",
        )
        Printer.print_text(
            "# Polyaxon completion\n. ~/.polyaxon-complete.zsh\n. ~/.plx-complete.zsh"
        )
        Printer.heading(
            "Reload your shell.",
        )
    elif shell == "fish":
        Printer.heading(
            "Please save the following scripts under `~/.config/fish/completions`:",
        )
        Printer.print(
            "_POLYAXON_COMPLETE=fish_source polyaxon > ~/.config/fish/completions/polyaxon-complete.fish"  # noqa
        )
        Printer.print(
            "_PLX_COMPLETE=fish_source plx > ~/.config/fish/completions/plx-complete.fish"
        )
    else:
        Printer.print("Shell {} is not supported.".format(shell))



import websocket

from clipped.utils.json import orjson_loads

from polyaxon.logger import logger


class SocketTransportMixin:
    """Socket operations transport."""

    def socket(self, url, message_handler, headers=None):
        webs = websocket.WebSocketApp(
            url,
            on_message=lambda ws, message: self._on_message(message_handler, message),
            on_error=self._on_error,
            on_close=self._on_close,
            header=self._get_headers(headers),
        )
        return webs

    def stream(self, url, message_handler, headers=None):
        webs = self.socket(url=url, message_handler=message_handler, headers=headers)
        webs.run_forever(ping_interval=30, ping_timeout=10)

    def _on_message(self, message_handler, message):
        if message_handler and message:
            if not isinstance(message, str):
                message = message.decode("utf-8")
            message_handler(orjson_loads(message))

    @staticmethod
    def _on_error(ws, error):
        if isinstance(error, (KeyboardInterrupt, SystemExit)):
            logger.info("Quitting... The session will be running in the background.")
        else:
            logger.debug("Termination cause: %s", error)
            logger.debug("Session disconnected.")

    @staticmethod
    def _on_close(ws):
        logger.info("Session ended")



from typing import List, Optional

from clipped.compact.pydantic import StrictStr
from clipped.config.schema import BaseAllowSchemaModel

from polyaxon._sdk.schemas.v1_team_member import V1TeamMember


class V1ListTeamMembersResponse(BaseAllowSchemaModel):
    count: Optional[int]
    results: Optional[List[V1TeamMember]]
    previous: Optional[StrictStr]
    next: Optional[StrictStr]



from uuid import UUID

from polyaxon._flow import V1EventKind, V1MatrixKind, V1RunKind


def get_fxt_templated_pipeline_without_params():
    return {
        "version": 1.1,
        "kind": "operation",
        "component": {
            "name": "test-build-run",
            "description": "testing a build and run pipeline",
            "tags": ["backend", "native"],
            "run": {
                "kind": V1RunKind.DAG,
                "operations": [
                    {
                        "dagRef": "build-template",
                        "name": "build",
                        "params": {
                            "env_vars": {
                                "value": [["env1", "value1"], ["env2", "value2"]]
                            }
                        },
                    },
                    {
                        "dagRef": "experiment-template",
                        "name": "run",
                        "dependencies": ["build"],
                        "params": {
                            "image": {
                                "ref": "ops.build",
                                "value": "outputs.docker-image",
                            },
                            "lr": {"value": 0.001},
                            "dag_uuid": {
                                "value": "globals.uuid",
                                "ref": "dag",
                                "contextOnly": True,
                            },
                            "op_uuid": {
                                "value": "globals.uuid",
                                "ref": "ops.build",
                                "contextOnly": True,
                            },
                        },
                    },
                ],
                "components": [
                    {
                        "name": "experiment-template",
                        "description": "experiment to predict something",
                        "tags": ["key", "value"],
                        "inputs": [
                            {
                                "name": "lr",
                                "type": "float",
                                "value": 0.1,
                                "isOptional": True,
                            },
                            {"name": "image", "type": "str"},
                        ],
                        "termination": {"maxRetries": 2},
                        "run": {
                            "kind": V1RunKind.JOB,
                            "environment": {
                                "nodeSelector": {"polyaxon": "experiments"},
                                "serviceAccountName": "service",
                                "imagePullSecrets": ["secret1", "secret2"],
                            },
                            "container": {
                                "image": "{{ image }}",
                                "command": ["python3", "main.py"],
                                "args": "--lr={{ lr }}",
                                "resources": {"requests": {"cpu": 1}},
                            },
                        },
                    },
                    {
                        "name": "build-template",
                        "description": "build images",
                        "tags": ["backend", "kaniko"],
                        "inputs": [
                            {"name": "env_vars", "type": "list", "isList": True}
                        ],
                        "outputs": [{"name": "docker-image", "type": "str"}],
                        "termination": {"maxRetries": 2},
                        "run": {
                            "kind": V1RunKind.JOB,
                            "environment": {
                                "nodeSelector": {"polyaxon": "experiments"},
                                "serviceAccountName": "service",
                                "imagePullSecrets": ["secret1", "secret2"],
                            },
                            "container": {
                                "image": "base",
                                "resources": {"requests": {"cpu": 1}},
                            },
                            "init": [
                                {
                                    "dockerfile": {
                                        "image": "base",
                                        "env": "{{ env_vars }}",
                                    }
                                }
                            ],
                        },
                    },
                ],
            },
        },
    }


def get_fxt_templated_pipeline_with_upstream_run(run_uuid: UUID):
    return {
        "version": 1.1,
        "kind": "operation",
        "component": {
            "name": "test-build-run",
            "description": "testing a build and run pipeline",
            "tags": ["backend", "native"],
            "inputs": [
                {
                    "name": "top",
                    "type": "int",
                    "value": 5,
                    "isOptional": True,
                    "description": "top jobs.",
                }
            ],
            "run": {
                "kind": V1RunKind.DAG,
                "operations": [
                    {
                        "dagRef": "build-template",
                        "name": "build",
                        "params": {
                            "env_vars": {
                                "value": [["env1", "value1"], ["env2", "value2"]]
                            }
                        },
                    },
                    {
                        "dagRef": "experiment-template",
                        "name": "run",
                        "dependencies": ["build"],
                        "runPatch": {
                            "environment": {
                                "nodeSelector": {
                                    "{{ dag_uuid }}": "node-{{ context_param }}",
                                }
                            }
                        },
                        "params": {
                            "image": {
                                "ref": "ops.build",
                                "value": "outputs.docker-image",
                            },
                            "lr": {"value": 0.001},
                            "some-run": {
                                "value": "outputs.some-int",
                                "ref": "runs.{}".format(run_uuid.hex),
                            },
                            "context_param": {
                                "value": "inputs.top",
                                "ref": "dag",
                                "contextOnly": True,
                            },
                            "dag_uuid": {
                                "value": "globals.uuid",
                                "ref": "dag",
                                "contextOnly": True,
                            },
                            "op_uuid": {
                                "value": "globals.uuid",
                                "ref": "ops.build",
                                "contextOnly": True,
                            },
                        },
                    },
                ],
                "components": [
                    {
                        "name": "experiment-template",
                        "description": "experiment to predict something",
                        "tags": ["key", "value"],
                        "inputs": [
                            {
                                "name": "lr",
                                "type": "float",
                                "value": 0.1,
                                "isOptional": True,
                            },
                            {"name": "image", "type": "str"},
                            {"name": "some-run", "type": "int"},
                        ],
                        "termination": {"maxRetries": 2},
                        "run": {
                            "kind": V1RunKind.JOB,
                            "environment": {
                                "nodeSelector": {"polyaxon": "experiments"},
                                "serviceAccountName": "service",
                                "imagePullSecrets": ["secret1", "secret2"],
                            },
                            "container": {
                                "image": "{{ image }}",
                                "command": ["python3", "main.py"],
                                "args": "--lr={{ lr }}",
                                "resources": {"requests": {"cpu": 1}},
                            },
                        },
                    },
                    {
                        "name": "build-template",
                        "description": "build images",
                        "tags": ["backend", "kaniko"],
                        "inputs": [
                            {"name": "env_vars", "type": "list", "isList": True}
                        ],
                        "outputs": [{"name": "docker-image", "type": "str"}],
                        "termination": {"maxRetries": 2},
                        "run": {
                            "kind": V1RunKind.JOB,
                            "environment": {
                                "nodeSelector": {"polyaxon": "experiments"},
                                "serviceAccountName": "service",
                                "imagePullSecrets": ["secret1", "secret2"],
                            },
                            "container": {
                                "image": "base",
                                "resources": {"requests": {"cpu": 1}},
                            },
                            "init": [
                                {
                                    "dockerfile": {
                                        "image": "base",
                                        "env": "{{ env_vars }}",
                                    }
                                }
                            ],
                        },
                    },
                ],
            },
        },
    }


def get_fxt_build_run_pipeline():
    return {
        "version": 1.1,
        "kind": "operation",
        "dependencies": ["foo", "bar"],
        "trigger": "all_succeeded",
        "component": {
            "name": "build_run_pipeline",
            "tags": ["foo", "bar"],
            "description": "testing a build and run pipeline",
            "run": {
                "kind": V1RunKind.DAG,
                "operations": [
                    {"dagRef": "build-template", "name": "A"},
                    {"dagRef": "job-template", "name": "B", "dependencies": ["A"]},
                ],
                "components": [
                    {
                        "name": "job-template",
                        "run": {"kind": V1RunKind.JOB, "container": {"image": "test"}},
                    },
                    {
                        "name": "build-template",
                        "tags": ["backend", "kaniko"],
                        "run": {
                            "kind": V1RunKind.JOB,
                            "container": {"image": "test"},
                            "init": [{"connection": "foo", "git": {"revision": "dev"}}],
                        },
                    },
                ],
            },
        },
    }


def get_fxt_train_tensorboard_events_pipeline():
    return {
        "version": 1.1,
        "kind": "operation",
        "dependencies": ["foo", "bar"],
        "trigger": "all_succeeded",
        "component": {
            "name": "train_tensorboard",
            "tags": ["foo", "bar"],
            "description": "testing an events driven pipeline",
            "run": {
                "kind": V1RunKind.DAG,
                "operations": [
                    {"dagRef": "train-template", "name": "A"},
                    {
                        "dagRef": "tensorboard-template",
                        "name": "B",
                        "events": [
                            {
                                "kinds": [V1EventKind.RUN_STATUS_RUNNING],
                                "ref": "ops.A",
                            }
                        ],
                    },
                ],
                "components": [
                    {
                        "name": "train-template",
                        "run": {"kind": V1RunKind.JOB, "container": {"image": "test"}},
                    },
                    {
                        "name": "tensorboard-template",
                        "run": {
                            "kind": V1RunKind.SERVICE,
                            "container": {"image": "test"},
                        },
                    },
                ],
            },
        },
    }


def get_fxt_build_run_pipeline_with_inputs():
    return {
        "version": 1.1,
        "kind": "operation",
        "dependencies": ["foo", "bar"],
        "params": {"param1": {"value": "foo"}, "param2": {"value": "bar"}},
        "trigger": "all_succeeded",
        "component": {
            "name": "my-pipe-test",
            "description": "testing a pipe",
            "tags": ["key", "value"],
            "inputs": [
                {"name": "param1", "type": "str"},
                {"name": "param2", "type": "str"},
            ],
            "run": {
                "kind": V1RunKind.DAG,
                "operations": [
                    {"dagRef": "build-template", "name": "A"},
                    {"dagRef": "job-template", "name": "B", "dependencies": ["A"]},
                ],
                "components": [
                    {
                        "name": "job-template",
                        "run": {"kind": V1RunKind.JOB, "container": {"image": "test"}},
                    },
                    {
                        "name": "build-template",
                        "tags": ["backend", "kaniko"],
                        "run": {
                            "kind": V1RunKind.JOB,
                            "container": {"image": "test"},
                            "init": [{"connection": "foo", "git": {"revision": "dev"}}],
                        },
                    },
                ],
            },
        },
    }


def get_fxt_pipeline_params_env_termination():
    return {
        "version": 1.1,
        "name": "params_env_termination",
        "kind": "operation",
        "dependencies": ["foo", "bar"],
        "trigger": "all_succeeded",
        "params": {"pipe_param1": {"value": "foo"}, "pipe_param2": {"value": "bar"}},
        "termination": {"maxRetries": 2},
        "runPatch": {
            "kind": V1RunKind.DAG,
            "environment": {
                "nodeSelector": {"polyaxon": "experiments"},
                "serviceAccountName": "service",
                "imagePullSecrets": ["secret1", "secret2"],
            },
        },
        "component": {
            "inputs": [
                {"name": "pipe_param1", "type": "str"},
                {"name": "pipe_param2", "type": "str"},
            ],
            "run": {
                "kind": V1RunKind.DAG,
                "operations": [
                    {
                        "dagRef": "build-template",
                        "name": "A",
                        "params": {
                            "param2": {"ref": "dag", "value": "inputs.pipe_param2"}
                        },
                    },
                    {
                        "dagRef": "job-template",
                        "name": "B",
                        "dependencies": ["A"],
                        "params": {
                            "param1": {"ref": "dag", "value": "inputs.pipe_param1"}
                        },
                        "termination": {"maxRetries": 3},
                        "runPatch": {"kind": V1RunKind.JOB},
                    },
                ],
                "components": [
                    {
                        "name": "job-template",
                        "inputs": [{"name": "param1", "type": "str"}],
                        "termination": {"maxRetries": 1},
                        "run": {"kind": V1RunKind.JOB, "container": {"image": "test"}},
                    },
                    {
                        "name": "build-template",
                        "tags": ["backend", "kaniko"],
                        "inputs": [{"name": "param2", "type": "str"}],
                        "termination": {"maxRetries": 1},
                        "run": {
                            "kind": V1RunKind.JOB,
                            "container": {"image": "test"},
                            "init": [{"connection": "foo", "git": {"revision": "dev"}}],
                        },
                    },
                ],
            },
        },
    }


def get_fxt_map_reduce():
    return {
        "version": 1.1,
        "kind": "operation",
        "dependencies": ["foo", "bar"],
        "params": {"image": {"value": "foo"}},
        "trigger": "all_succeeded",
        "component": {
            "name": "map-reduce",
            "description": "testing a pipe",
            "tags": ["key", "value"],
            "inputs": [
                {"name": "image", "type": "str"},
            ],
            "run": {
                "kind": V1RunKind.DAG,
                "operations": [
                    {
                        "dagRef": "job-template",
                        "name": "A",
                        "params": {"image": {"ref": "dag", "value": "inputs.image"}},
                        "matrix": {
                            "kind": V1MatrixKind.MAPPING,
                            "values": [
                                {"param1": "test1", "param2": 1, "param3": 1.1},
                                {"param1": "test2", "param2": 2, "param3": 2.1},
                            ],
                        },
                    },
                    {
                        "dagRef": "reduce-template",
                        "name": "B",
                        "dependencies": ["A"],
                        "params": {
                            "map_controller": {
                                "ref": "ops.A",
                                "value": "globals.uuid",
                                "contextOnly": True,
                            },
                        },
                        "joins": [
                            {
                                "query": "pipeline: {{ map_controller }}",
                                "params": {
                                    "metrics": {"value": "outputs.metrics"},
                                    "run_metrics_events": {
                                        "value": "artifacts.metrics",
                                        "contextOnly": True,
                                    },
                                    "run_outputs": {
                                        "value": "artifacts.outputs",
                                        "contextOnly": True,
                                    },
                                    "files": {
                                        "value": {
                                            "files": [
                                                "subpath/file1",
                                                "different/subpath/file2",
                                            ],
                                        },
                                        "contextOnly": True,
                                    },
                                },
                            },
                        ],
                    },
                ],
                "components": [
                    {
                        "name": "job-template",
                        "inputs": [
                            {
                                "name": "lr",
                                "type": "float",
                                "value": 0.1,
                                "isOptional": True,
                            },
                            {"name": "image", "type": "str"},
                            {"name": "param1", "type": "str"},
                            {"name": "param2", "type": "int"},
                            {"name": "param3", "type": "float"},
                        ],
                        "outputs": [
                            {"name": "result1", "type": "float"},
                        ],
                        "run": {
                            "kind": V1RunKind.JOB,
                            "environment": {
                                "nodeSelector": {"polyaxon": "experiments"},
                                "serviceAccountName": "service",
                                "imagePullSecrets": ["secret1", "secret2"],
                            },
                            "container": {
                                "image": "{{ image }}",
                                "command": ["python3", "main.py"],
                                "args": "--lr={{ lr }}",
                                "resources": {"requests": {"cpu": 1}},
                            },
                        },
                    },
                    {
                        "name": "reduce-template",
                        "inputs": [
                            {"name": "metrics", "type": "float", "isList": True},
                        ],
                        "run": {
                            "kind": V1RunKind.JOB,
                            "container": {"image": "test"},
                        },
                    },
                ],
            },
        },
    }



from typing import Dict, List, Optional

from clipped.utils.lists import to_list

from polyaxon._connections import V1Connection
from polyaxon._flow import V1Init


def get_connection_annotations(
    artifacts_store: Optional[V1Connection],
    init_connections: Optional[List[V1Init]],
    connections: List[str],
    connection_by_names: Optional[Dict[str, V1Connection]],
) -> Dict:
    """Resolve all annotations to inject per replica"""
    connections = to_list(connections, check_none=True)
    init_connections = to_list(init_connections, check_none=True)
    connection_by_names = connection_by_names or {}

    requested_connection_names = connections[:]
    for init_connection in init_connections:
        if (
            init_connection.connection
            and init_connection.connection not in requested_connection_names
        ):
            requested_connection_names.append(init_connection.connection)
    if artifacts_store and artifacts_store.name not in requested_connection_names:
        requested_connection_names.append(artifacts_store.name)

    annotations = {}
    for c in requested_connection_names:
        annotations.update(connection_by_names[c].annotations or {})

    return annotations



import datetime

from typing import Any, Dict, List, Optional

from clipped.compact.pydantic import StrictStr
from clipped.config.schema import BaseAllowSchemaModel
from clipped.types.uuids import UUIDStr

from polyaxon._schemas.lifecycle import V1Statuses


class V1Agent(BaseAllowSchemaModel):
    uuid: Optional[UUIDStr]
    name: Optional[StrictStr]
    description: Optional[StrictStr]
    tags: Optional[List[StrictStr]]
    live_state: Optional[int]
    namespace: Optional[StrictStr]
    version_api: Optional[Dict[str, Any]]
    version: Optional[StrictStr]
    content: Optional[StrictStr]
    created_at: Optional[datetime.datetime]
    updated_at: Optional[datetime.datetime]
    status: Optional[V1Statuses]
    is_replica: Optional[bool]
    is_ui_managed: Optional[bool]
    settings: Optional[Dict[str, Any]]



import copy

from typing import Dict, Iterable, List, Optional

from clipped.utils.sanitizers import sanitize_string_dict
from clipped.utils.strings import slugify

from polyaxon import pkg, settings
from polyaxon._connections import V1Connection, V1ConnectionResource
from polyaxon._flow import V1Environment, V1Init, V1Plugins
from polyaxon._k8s import k8s_schemas
from polyaxon._k8s.converter.base.containers import ContainerMixin
from polyaxon._k8s.converter.base.env_vars import EnvMixin
from polyaxon._k8s.converter.base.init import InitConverter
from polyaxon._k8s.converter.base.main import MainConverter
from polyaxon._k8s.converter.base.mounts import MountsMixin
from polyaxon._k8s.converter.base.sidecar import SidecarConverter
from polyaxon._k8s.converter.common.annotations import get_connection_annotations
from polyaxon._k8s.converter.pod.volumes import get_pod_volumes
from polyaxon._k8s.replica import ReplicaSpec
from polyaxon._runner.converter import BaseConverter as _BaseConverter
from polyaxon._runner.kinds import RunnerKind
from polyaxon.exceptions import PolyaxonConverterError


class BaseConverter(
    MainConverter,
    InitConverter,
    SidecarConverter,
    ContainerMixin,
    EnvMixin,
    MountsMixin,
    _BaseConverter,
):
    RUNNER_KIND = RunnerKind.K8S
    GROUP: Optional[str] = None
    API_VERSION: Optional[str] = None
    PLURAL: Optional[str] = None
    K8S_ANNOTATIONS_KIND: Optional[str] = None
    K8S_LABELS_COMPONENT: Optional[str] = None
    K8S_LABELS_PART_OF: Optional[str] = None

    def is_valid(self):
        super().is_valid()
        if not self.GROUP:
            raise PolyaxonConverterError(
                "Please make sure that a converter subclass has a valid GROUP"
            )
        if not self.API_VERSION:
            raise PolyaxonConverterError(
                "Please make sure that a converter subclass has a valid API_VERSION"
            )
        if not self.PLURAL:
            raise PolyaxonConverterError(
                "Please make sure that a converter subclass has a valid PLURAL"
            )
        if not self.K8S_ANNOTATIONS_KIND:
            raise PolyaxonConverterError(
                "Please make sure that a converter subclass has a valid K8S_ANNOTATIONS_KIND"
            )
        if not self.K8S_LABELS_COMPONENT:
            raise PolyaxonConverterError(
                "Please make sure that a converter subclass has a valid K8S_LABELS_COMPONENT"
            )
        if not self.K8S_LABELS_PART_OF:
            raise PolyaxonConverterError(
                "Please make sure that a converter subclass has a valid K8S_LABELS_PART_OF"
            )

    def get_recommended_labels(self, version: str) -> Dict:
        return {
            "app.kubernetes.io/name": slugify(
                self.run_name[:63] if self.run_name else self.run_name
            ),
            "app.kubernetes.io/instance": self.run_uuid,
            "app.kubernetes.io/version": version,
            "app.kubernetes.io/part-of": self.K8S_LABELS_PART_OF,
            "app.kubernetes.io/component": self.K8S_LABELS_COMPONENT,
            "app.kubernetes.io/managed-by": "polyaxon",
        }

    @property
    def annotations(self) -> Dict:
        return {
            "operation.polyaxon.com/name": self.run_name,
            "operation.polyaxon.com/owner": self.owner_name,
            "operation.polyaxon.com/project": self.project_name,
            "operation.polyaxon.com/kind": self.K8S_ANNOTATIONS_KIND,
        }

    def get_annotations(
        self,
        annotations: Dict,
        artifacts_store: Optional[V1Connection],
        init_connections: Optional[List[V1Init]],
        connections: List[str],
        connection_by_names: Optional[Dict[str, V1Connection]],
    ) -> Dict:
        annotations = annotations or {}
        annotations = copy.copy(annotations)
        connections_annotations = get_connection_annotations(
            artifacts_store=artifacts_store,
            init_connections=init_connections,
            connections=connections,
            connection_by_names=connection_by_names,
        )
        connections_annotations = connections_annotations or {}
        annotations.update(connections_annotations)
        annotations.update(self.annotations)
        return sanitize_string_dict(annotations)

    def get_labels(self, version: str, labels: Dict) -> Dict:
        labels = labels or {}
        labels = copy.copy(labels)
        labels.update(self.get_recommended_labels(version=version))
        return sanitize_string_dict(labels)

    @staticmethod
    def _new_container(name: str) -> k8s_schemas.V1Container:
        return k8s_schemas.V1Container(name=name)

    @classmethod
    def _ensure_container(
        cls, container: k8s_schemas, volumes: List[k8s_schemas.V1Volume]
    ) -> k8s_schemas.V1Container:
        return container

    def get_replica_resource(
        self,
        environment: V1Environment,
        plugins: V1Plugins,
        volumes: List[k8s_schemas.V1Volume],
        init: List[V1Init],
        sidecars: List[k8s_schemas.V1Container],
        container: k8s_schemas.V1Container,
        artifacts_store: V1Connection,
        connections: List[str],
        connection_by_names: Dict[str, V1Connection],
        secrets: Optional[Iterable[V1ConnectionResource]],
        config_maps: Optional[Iterable[V1ConnectionResource]],
        kv_env_vars: List[List],
        default_sa: Optional[str] = None,
        ports: List[int] = None,
        num_replicas: Optional[int] = None,
        custom: Dict = None,
    ) -> ReplicaSpec:
        volumes = volumes or []
        init = init or []
        sidecars = sidecars or []
        connections = connections or []
        environment = environment or V1Environment()
        environment.service_account_name = (
            environment.service_account_name
            or default_sa
            or settings.AGENT_CONFIG.runs_sa
        )

        init_connections = self.filter_connections_from_init(init=init)

        volumes = get_pod_volumes(
            plugins=plugins,
            artifacts_store=artifacts_store,
            init_connections=init_connections,
            connections=connections,
            connection_by_names=connection_by_names,
            secrets=secrets,
            config_maps=config_maps,
            volumes=volumes,
        )

        init_containers = self.get_init_containers(
            polyaxon_init=self.polyaxon_init,
            plugins=plugins,
            artifacts_store=artifacts_store,
            init_connections=init_connections,
            init_containers=self.filter_containers_from_init(init=init),
            connection_by_names=connection_by_names,
            log_level=plugins.log_level,
        )

        sidecar_containers = self.get_sidecar_containers(
            polyaxon_sidecar=self.polyaxon_sidecar,
            plugins=plugins,
            artifacts_store=artifacts_store,
            sidecar_containers=sidecars,
            log_level=plugins.log_level,
        )

        main_container = self.get_main_container(
            main_container=container,
            plugins=plugins,
            artifacts_store=artifacts_store,
            connections=connections,
            init_connections=init_connections,
            connection_by_names=connection_by_names,
            secrets=secrets,
            config_maps=config_maps,
            ports=ports,
            kv_env_vars=kv_env_vars,
        )

        labels = self.get_labels(version=pkg.VERSION, labels=environment.labels)
        annotations = self.get_annotations(
            annotations=environment.annotations,
            artifacts_store=artifacts_store,
            connections=connections,
            init_connections=init_connections,
            connection_by_names=connection_by_names,
        )
        return ReplicaSpec(
            volumes=volumes,
            init_containers=[self._sanitize_container(c) for c in init_containers],
            sidecar_containers=[
                self._sanitize_container(c) for c in sidecar_containers
            ],
            main_container=main_container,
            labels=labels,
            annotations=annotations,
            environment=environment,
            num_replicas=num_replicas,
            custom=custom,
        )



from typing import Optional

from clipped.compact.pydantic import Field, StrictInt, StrictStr

from polyaxon._schemas.base import BaseSchemaModel


class EmailConfig(BaseSchemaModel):
    enabled: Optional[bool]
    email_from: Optional[StrictStr] = Field(alias="from")
    host: Optional[StrictStr]
    port: Optional[StrictInt]
    use_tls: Optional[bool] = Field(alias="useTls")
    host_user: Optional[StrictStr] = Field(alias="hostUser")
    host_password: Optional[StrictStr] = Field(alias="hostPassword")
    backend: Optional[StrictStr]



from polyaxon._utils.fixtures.bo import (
    get_fxt_bo_with_inputs_outputs,
    get_fxt_bo_with_run_patch,
)
from polyaxon._utils.fixtures.build import set_build_fixture
from polyaxon._utils.fixtures.grid import get_fxt_grid_with_inputs_outputs
from polyaxon._utils.fixtures.jobs import (
    get_fxt_job,
    get_fxt_job_with_inputs,
    get_fxt_job_with_inputs_and_conditions,
    get_fxt_job_with_inputs_and_joins,
    get_fxt_job_with_inputs_outputs,
    get_fxt_ray_job,
    get_fxt_tf_job,
)
from polyaxon._utils.fixtures.mapping import (
    get_fxt_mapping_with_inputs_outputs,
    get_fxt_mapping_with_run_patch,
)
from polyaxon._utils.fixtures.pipelines import (
    get_fxt_build_run_pipeline,
    get_fxt_build_run_pipeline_with_inputs,
    get_fxt_map_reduce,
    get_fxt_pipeline_params_env_termination,
    get_fxt_templated_pipeline_with_upstream_run,
    get_fxt_templated_pipeline_without_params,
    get_fxt_train_tensorboard_events_pipeline,
)
from polyaxon._utils.fixtures.schedule import get_fxt_schedule_with_inputs_outputs
from polyaxon._utils.fixtures.services import (
    get_fxt_job_with_hub_ref,
    get_fxt_service,
    get_fxt_service_with_inputs,
    get_fxt_service_with_upstream_runs,
)



from typing import Optional

from clipped.compact.pydantic import StrictStr
from clipped.config.schema import BaseAllowSchemaModel
from clipped.types.uuids import UUIDStr


class V1RunConnection(BaseAllowSchemaModel):
    uuid: Optional[UUIDStr]
    name: Optional[StrictStr]
    kind: Optional[StrictStr]



from typing import Any, Dict, List, Optional, Union

from clipped.compact.pydantic import Field, StrictStr
from clipped.config.patch_strategy import PatchStrategy
from clipped.types.ref_or_obj import RefField

from polyaxon._flow.cache import V1Cache
from polyaxon._flow.params import V1Param
from polyaxon._schemas.base import BaseSchemaModel


class V1Build(BaseSchemaModel):
    """You can configure Polyaxon to automatically trigger a build process anytime
    the component or operation is instantiated.

    the build section allows to dynamically recreate a new docker image
    that will be used to run the main logic.
    The new image is generated automatically and automatically set on the main container,
    the name is based on the project name and the run's uuid, i.e. `project:build-uuid`.

    > **Note**: When the build and matrix sections are used together,
    > a single build operation will be scheduled and will be used for all runs.

    > **Note**: When the build section is defined and an upload is triggered,
    > the uploaded artifacts will be pushed to the build run.

    Args:
        hub_ref: str
        connection: str
        queue: str, optional
        presets: List[str], optional
        cache: [V1Cache](/docs/automation/helpers/cache/), optional
        params: Dict[str, [V1Param](/docs/core/specification/params/)], optional
        run_patch: Dict, optional
        patch_strategy: str, optional, defaults to post_merge

    ## YAML usage

    ```yaml
    >>> build:
    >>>   hubRef: kaniko
    >>>   connection: registry-connection-name
    >>>   params:
    >>>     context:
    >>>       value: "path/to/context"
    ```

    ## Python usage

    ```python
    >>> from polyaxon.schemas import V1Statuses, V1Build, V1Param
    >>> build = V1Build(
    >>>     hub_ref="kaniko",
    >>>     connection="registry-connection-name",
    >>>     params={
    >>>         "context": V1Param(value="path/to/context"),
    >>>         ...
    >>>     },
    >>> )
    ```

    ## Fields

    ### hubRef

    Polyaxon provides a [Component Hub](/docs/management/component-hub/)
    for hosting versioned components with an access control system to improve
    the productivity of your team.

    To trigger a build based on a component hosted on Polyaxon Component Hub, you can use `hubRef`

    ```yaml
    >>> build:
    >>>   hubRef: kaniko
    ...
    ```

    Or custom hook component

    ```yaml
    >>> build:
    >>>   hubRef: my-component:dev
    ...
    ```

    ### Connection

    The connection to use for pushing the image. Polyaxon will automatically generate a valid image
    based on the project name and the build operation uuid.

    ```yaml
    >>> build:
    >>>   connection: registry-conneciton-name
    ...
    ```

    ### queue

    The [queue](/docs/core/scheduling-strategies/queues/) to use.
    If not provided, the default queue will be used.

    ```yaml
    >>> build:
    >>>   queue: agent-name/queue-name
    ```

    If the agent name is not specified, Polyaxon will resolve the name of the queue
    based on the default agent.

    ```yaml
    >>> build:
    >>>   queue: queue-name
    ```

    ### presets

    The [presets](/docs/management/organizations/presets/) to use for the hook operation,
    if provided, it will override the component's presets otherwise
    the presets of the component will be used if available.

    ```yaml
    >>> build:
    >>>   presets: [test]
    ```

    ### cache

    The [cache](/docs/automation/helpers/cache/) to use for the build operation,
    if provided, it will override the component's cache otherwise
    the cache of the component will be used if it exists.

    ```yaml
    >>> operation:
    >>>   cache:
    >>>     disable: false
    >>>     ttl: 100
    ```

    ### params

    The [params](/docs/core/specification/params/) to pass if the handler requires extra params,
    they will be validated against the inputs/outputs.
    If a parameter is passed and the component does not define a corresponding inputs/outputs,
    a validation error will be raised unless the param has the `contextOnly` flag enabled.

    ```yaml
    >>> build:
    >>>   params:
    >>>     param1: {value: 1.1}
    >>>     param2: {value: test}
    >>>   ...
    ```

    Params can be used to pass any number of parameters that the component is expecting.

    Example setting the `context`:

    ```yaml
    >>> build:
    >>>   params:
    >>>     context:
    >>>       value: "{{ globals.artifacts_path }}/repo-name"
    ```

    You can use the params field to define your own image destination with either a
    fixed value like `foo/bar:test` or
    templated with the context information:

    ```yaml
    >>> build:
    >>>   params:
    >>>     destination:
    >>>       value: "org/repo:{{ globals.uuid }}"
    ```

    Or provide the connection under the parameter

    ```yaml
    >>> build:
    >>>   params:
    >>>     destination:
    >>>       value: "org/{{ globals.project_name }}:{{ globals.uuid }}"
    >>>       connection: connection-name
    ```

    Polyaxon will use by default the following destination if no destination parameter is provided:
    `{{ globals.project_name }}:{{ globals.uuid }}` and it will use by default the connection
    defined in the build process.

    ### runPatch

    The run patch provides a way to override information about the component's run section,
    for example the container's resources, the environment section, or the init section.

    The run patch is a dictionary that can modify most of the runtime information and
    will be resolved against the corresponding run kind, in this case
    [V1Job](/docs/experimentation/jobs/).

    ### patchStrategy

    Defines how the compiler should handle keys that are defined on the component,
    or how to merge multiple presets when using the override behavior `-f`.

    There are four strategies:
     * `replace`: replaces all keys with new values if provided.
     * `isnull`: only applies new values if the keys have empty/None values.
     * `post_merge`: applies deep merge where newer values are applied last.
     * `pre_merge`: applies deep merge where newer values are applied first.
    """

    _IDENTIFIER = "build"
    hub_ref: StrictStr = Field(alias="hubRef")
    connection: Optional[StrictStr]
    queue: Optional[StrictStr]
    presets: Optional[Union[List[StrictStr], RefField]]
    cache: Optional[Union[V1Cache, RefField]]
    params: Optional[Dict[str, Union[V1Param, RefField]]]
    run_patch: Optional[Dict[str, Any]] = Field(alias="runPatch")
    patch_strategy: Optional[Union[PatchStrategy, RefField]] = Field(
        alias="patchStrategy"
    )



import os

from collections import OrderedDict
from typing import Dict, List, Optional, Union

from clipped.formatting import Printer
from clipped.utils.lists import to_list

from polyaxon._cli.errors import handle_cli_error
from polyaxon._config.spec import ConfigSpec
from polyaxon._flow import V1Dag, V1Init, V1Matrix, V1Operation
from polyaxon._polyaxonfile.manager import get_op_specification
from polyaxon._polyaxonfile.params import parse_hparams, parse_params
from polyaxon._polyaxonfile.specs import get_specification, kinds
from polyaxon.exceptions import PolyaxonfileError, PolyaxonSchemaError


def collect_dag_components(dag: V1Dag, path_context: Optional[str] = None):
    """Collect components that cannot be resolved by the scheduler"""
    for op in dag.operations:
        op_name = op.name
        if op.has_url_reference or op.has_path_reference or op.has_hub_reference:
            try:
                op = collect_references(op, path_context)
            except Exception as e:
                raise PolyaxonSchemaError(
                    "Pipeline op with name `{}` requires a component with ref `{}`, "
                    "the reference could not be resolved. Error: {}".format(
                        op_name, op.hub_ref or op.url_ref or op.path_ref, e
                    )
                )


def collect_references(config: V1Operation, path_context: Optional[str] = None):
    if config.has_component_reference:
        return config
    elif config.has_hub_reference:
        component = ConfigSpec.get_from(config.hub_ref, "hub").read()
    elif config.has_url_reference:
        component = ConfigSpec.get_from(config.url_ref, "url").read()
    elif config.has_path_reference:
        path_ref = config.path_ref
        if path_context:
            path_ref = os.path.join(
                os.path.dirname(os.path.abspath(path_context)), path_ref
            )
        component = ConfigSpec.get_from(path_ref).read()
        path_context = path_ref
    else:
        raise PolyaxonfileError("Operation found without component")
    component = get_specification(data=component)
    if component.kind != kinds.COMPONENT:
        if config.has_url_reference:
            ref_type = "Url ref"
            ref = config.url_ref
        else:
            ref_type = "Path ref"
            ref = config.path_ref
        raise PolyaxonfileError(
            "the reference ({}) `{}` is of kind `{}`, it should be a `{}`".format(
                ref, ref_type, component.kind, kinds.COMPONENT
            )
        )
    config.component = component
    if component.is_dag_run:
        collect_dag_components(component.run, path_context)
    return config


def check_polyaxonfile(
    polyaxonfile: Optional[str] = None,
    python_module: Optional[str] = None,
    url: Optional[str] = None,
    hub: Optional[str] = None,
    params: Optional[Union[List[str], Dict]] = None,
    hparams: Optional[Union[List[str], Dict]] = None,
    matrix_kind: Optional[str] = None,
    matrix_concurrency: Optional[int] = None,
    matrix_num_runs: Optional[int] = None,
    matrix: Optional[Union[Dict, V1Matrix]] = None,
    presets: Optional[List[str]] = None,
    queue: Optional[str] = None,
    nocache: Optional[bool] = None,
    cache: Optional[Union[int, str, bool]] = None,
    verbose: bool = True,
    is_cli: bool = True,
    to_op: bool = True,
    validate_params: bool = True,
    approved: Optional[Union[int, str, bool]] = None,
    git_init: Optional[V1Init] = None,
    ignore_template: bool = False,
):
    if sum([1 for i in [python_module, url, hub] if i]) > 1:
        message = (
            "You can only use one and only one option: "
            "hub, url, or a python module.".format(hub)
        )
        if is_cli:
            Printer.error(message, sys_exit=True)
        else:
            raise PolyaxonfileError(message)
    if not any([polyaxonfile, python_module, url, hub]):
        polyaxonfile = check_default_path(path=".")
    if not any([polyaxonfile, python_module, url, hub]):
        message = (
            "Something went wrong, `check_polyaxonfile` was called without a polyaxonfile, "
            "a hub component reference, a url or a python module."
        )
        if is_cli:
            Printer.error(message, sys_exit=True)
        else:
            raise PolyaxonfileError(message)
    if hub and not to_op:
        message = "Something went wrong, calling hub component `{}` without operation.".format(
            hub
        )
        if is_cli:
            Printer.error(message, sys_exit=True)
        else:
            raise PolyaxonfileError(message)

    polyaxonfile = to_list(polyaxonfile, check_none=True)

    parsed_params = None
    if params:
        parsed_params = parse_params(params, is_cli=is_cli)
    parsed_hparams = None
    if hparams:
        parsed_hparams = parse_hparams(hparams, is_cli=is_cli)

    if not any([os.path.isfile(f) for f in polyaxonfile]) and not any(
        [python_module, url, hub]
    ):
        message = "Please pass a valid polyaxonfile, a python module, a url, or a component name"
        if is_cli:
            Printer.error(message, sys_exit=True)
        else:
            raise PolyaxonfileError(message)

    try:
        path_context = None

        if python_module:
            path_context = python_module
            plx_file = (
                ConfigSpec.get_from(python_module, config_type=".py")
                .read()
                .to_dict(include_kind=True, include_version=True)
            )
        elif url:
            plx_file = ConfigSpec.get_from(url, "url").read()
        elif hub:
            plx_file = ConfigSpec.get_from(hub, "hub").read()
        else:
            path_context = polyaxonfile.pop(0)
            plx_file = ConfigSpec.read_from(path_context)

        plx_file = get_specification(data=plx_file)
        if plx_file.kind == kinds.OPERATION:
            plx_file = collect_references(plx_file, path_context)
            plx_component = plx_file.component
        else:
            plx_component = plx_file

        if plx_component.is_dag_run:
            collect_dag_components(plx_component.run, path_context)

        if to_op or hub:
            plx_file = get_op_specification(
                hub=hub,
                config=plx_file,
                params=parsed_params,
                hparams=parsed_hparams,
                matrix_kind=matrix_kind,
                matrix_concurrency=matrix_concurrency,
                matrix_num_runs=matrix_num_runs,
                matrix=matrix,
                presets=presets,
                queue=queue,
                nocache=nocache,
                cache=cache,
                approved=approved,
                validate_params=validate_params,
                preset_files=polyaxonfile,
                git_init=git_init,
            )
        if verbose and is_cli:
            Printer.success("Polyaxonfile valid")
        if ignore_template:
            plx_file.disable_template()
        if plx_file.is_template():
            template_message = "This polyaxonfile was marked as template by the owner:"
            if plx_file.template.description:
                template_message += "\ntemplate description: {}".format(
                    plx_file.template.description
                )
            if plx_file.template.fields:
                template_message += "\ntemplate fields that need changes: {}".format(
                    plx_file.template.fields
                )
            Printer.warning(template_message)
        return plx_file
    except Exception as e:
        message = "Polyaxonfile is not valid."
        if is_cli:
            handle_cli_error(e, message=message, sys_exit=True)
        else:
            raise PolyaxonfileError(message) from e


def check_polyaxonfile_kind(specification, kind):
    if specification.kind != kind:
        Printer.error(
            "Your polyaxonfile must be of kind: `{}`, "
            "received: `{}`.".format(kind, specification.kind),
            sys_exit=True,
        )


def get_matrix_info(kind, concurrency, early_stopping=False, **kwargs):
    info = OrderedDict()
    info["Matrix kind"] = kind.lower()
    info["Concurrency"] = (
        "{} runs".format("sequential")
        if concurrency == 1
        else "{} concurrent runs".format(concurrency)
    )
    info["Early stopping"] = "activated" if early_stopping else "deactivated"
    if "num_runs" in kwargs:
        info["Num of runs to create"] = kwargs["num_runs"]

    Printer.dict_tabulate(info)


DEFAULT_POLYAXON_FILE_NAME = [
    "polyaxon",
    "polyaxonci",
    "polyaxon-ci",
    "polyaxon.ci",
    "polyaxonfile",
]

DEFAULT_POLYAXON_FILE_EXTENSION = ["yaml", "yml", "json"]


def check_default_path(path):
    path = os.path.abspath(path)
    for filename in DEFAULT_POLYAXON_FILE_NAME:
        for ext in DEFAULT_POLYAXON_FILE_EXTENSION:
            filepath = os.path.join(path, "{}.{}".format(filename, ext))
            if os.path.isfile(filepath):
                return filepath



import datetime

from typing import Any, Dict, List, Optional

from clipped.compact.pydantic import StrictStr
from clipped.config.schema import BaseAllowSchemaModel
from clipped.types.uuids import UUIDStr

from polyaxon._schemas.lifecycle import V1ProjectVersionKind, V1StageCondition, V1Stages


class V1ProjectVersion(BaseAllowSchemaModel):
    uuid: Optional[UUIDStr]
    name: Optional[StrictStr]
    description: Optional[StrictStr]
    tags: Optional[List[StrictStr]]
    owner: Optional[StrictStr]
    project: Optional[StrictStr]
    connection: Optional[StrictStr]
    run: Optional[StrictStr]
    artifacts: Optional[List[StrictStr]]
    meta_info: Optional[Dict[str, Any]]
    created_at: Optional[datetime.datetime]
    updated_at: Optional[datetime.datetime]
    stage: Optional[V1Stages]
    kind: Optional[V1ProjectVersionKind]
    stage_conditions: Optional[List[V1StageCondition]]
    content: Optional[StrictStr]
    readme: Optional[StrictStr]
    state: Optional[StrictStr]
    role: Optional[StrictStr]
    contributors: Optional[Dict[str, Any]]



import datetime

from typing import List, Optional

from clipped.compact.pydantic import StrictStr
from clipped.config.schema import BaseAllowSchemaModel
from clipped.types.uuids import UUIDStr

from polyaxon._sdk.schemas.v1_team_settings import V1TeamSettings


class V1Team(BaseAllowSchemaModel):
    uuid: Optional[UUIDStr]
    name: Optional[StrictStr]
    projects: Optional[List[StrictStr]]
    component_hubs: Optional[List[StrictStr]]
    model_registries: Optional[List[StrictStr]]
    settings: Optional[V1TeamSettings]
    created_at: Optional[datetime.datetime]
    updated_at: Optional[datetime.datetime]



from typing import Optional

from clipped.compact.pydantic import StrictStr
from clipped.config.schema import BaseAllowSchemaModel


class V1AnalyticsSpec(BaseAllowSchemaModel):
    view: Optional[StrictStr]
    trunc: Optional[StrictStr]
    groupby: Optional[StrictStr]
    frequency: Optional[StrictStr]



import traceback

from typing import Optional

from polyaxon._schemas.lifecycle import V1StatusCondition, V1Statuses
from polyaxon.client import PolyaxonClient, V1Agent, V1AgentStateResponse
from polyaxon.logger import logger


class AgentClient:
    def __init__(
        self,
        owner: Optional[str] = None,
        agent_uuid: Optional[str] = None,
        client: Optional[PolyaxonClient] = None,
        is_async: bool = False,
    ):
        self.owner = owner
        self.agent_uuid = agent_uuid
        self.is_async = is_async
        self._client = client

    @property
    def client(self):
        if self._client:
            return self._client
        self._client = PolyaxonClient(is_async=self.is_async)
        return self._client

    @property
    def _is_managed(self):
        return self.owner is not None and self.agent_uuid is not None

    def get_info(self) -> V1Agent:
        return self.client.agents_v1.get_agent(owner=self.owner, uuid=self.agent_uuid)

    def get_state(self) -> V1AgentStateResponse:
        if self._is_managed:
            return self.client.agents_v1.get_agent_state(
                owner=self.owner, uuid=self.agent_uuid
            )
        return self.client.agents_v1.get_global_state(owner=self.owner)

    def log_agent_status(
        self, status: str, reason: Optional[str] = None, message: Optional[str] = None
    ):
        if not self._is_managed:
            return
        status_condition = V1StatusCondition.get_condition(
            type=status, status=True, reason=reason, message=message
        )
        return self.client.agents_v1.create_agent_status(
            owner=self.owner,
            uuid=self.agent_uuid,
            body={"condition": status_condition},
        )

    def sync_agent(self, agent: V1Agent):
        return self.client.agents_v1.sync_agent(
            owner=self.owner,
            agent_uuid=self.agent_uuid,
            body=agent,
        )

    def cron_agent(self):
        return self.client.agents_v1.cron_agent(owner=self.owner, _request_timeout=10)

    def log_agent_running(self):
        return self.log_agent_status(status=V1Statuses.RUNNING, reason="AgentLogger")

    def log_agent_failed(self, message=None):
        return self.log_agent_status(
            status=V1Statuses.FAILED, reason="AgentLogger", message=message
        )

    def log_agent_warning(self):
        return self.log_agent_status(
            status=V1Statuses.WARNING,
            reason="AgentLogger",
            message="The agent was interrupted, please check your deployment.",
        )

    def log_run_failed(
        self,
        run_owner: str,
        run_project: str,
        run_uuid: str,
        exc: Exception,
        message: Optional[str] = None,
    ):
        message = message or "Agent failed deploying run.\n"
        message += "error: {}\n{}".format(repr(exc), traceback.format_exc())
        logger.warning(message)
        return self.log_run_status(
            run_owner=run_owner,
            run_project=run_project,
            run_uuid=run_uuid,
            status=V1Statuses.FAILED,
            reason="AgentLogger",
            message=message,
        )

    def log_run_stopped(self, run_owner: str, run_project: str, run_uuid: str):
        message = "Run was not found. The agent assumed it was already stopped."
        logger.warning(message)
        return self.log_run_status(
            run_owner=run_owner,
            run_project=run_project,
            run_uuid=run_uuid,
            status=V1Statuses.STOPPED,
            reason="AgentLogger",
            message=message,
        )

    def log_run_scheduled(self, run_owner: str, run_project: str, run_uuid: str):
        message = "Run was scheduled by the agent."
        logger.info(message)
        return self.log_run_status(
            run_owner=run_owner,
            run_project=run_project,
            run_uuid=run_uuid,
            status=V1Statuses.SCHEDULED,
            reason="AgentLogger",
            message=message,
        )

    def log_run_running(self, run_owner: str, run_project: str, run_uuid: str):
        message = "Run changes were applied by the agent."
        logger.info(message)
        return self.log_run_status(
            run_owner=run_owner,
            run_project=run_project,
            run_uuid=run_uuid,
            status=V1Statuses.RUNNING,
            reason="AgentLogger",
            message=message,
        )

    def log_run_status(
        self,
        run_owner: str,
        run_project: str,
        run_uuid: str,
        status: str,
        reason: Optional[str] = None,
        message: Optional[str] = None,
    ):
        status_condition = V1StatusCondition.get_condition(
            type=status, status=True, reason=reason, message=message
        )
        return self.client.runs_v1.create_run_status(
            owner=run_owner,
            project=run_project,
            uuid=run_uuid,
            body={"condition": status_condition},
        )



import os

from typing import Type

from polyaxon._config.manager import ConfigManager
from polyaxon._config.spec import ConfigSpec
from polyaxon._contexts import paths as ctx_paths
from polyaxon._schemas.authentication import AccessTokenConfig


class AuthConfigManager(ConfigManager):
    """Manages access token configuration .auth file."""

    VISIBILITY = ConfigManager.Visibility.GLOBAL
    CONFIG_FILE_NAME = ".auth"
    CONFIG: Type[AccessTokenConfig] = AccessTokenConfig

    @classmethod
    def get_config_from_env(cls) -> AccessTokenConfig:
        tmp_path = cls.get_tmp_config_path()
        user_path = cls.get_global_config_path()

        auth_config = cls._CONFIG_READER.read_configs(
            [
                ConfigSpec(tmp_path, config_type=".json", check_if_exists=False),
                ConfigSpec(user_path, config_type=".json", check_if_exists=False),
                os.environ,
                ConfigSpec(
                    ctx_paths.CONTEXT_MOUNT_AUTH,
                    config_type=".json",
                    check_if_exists=False,
                ),
                {"dummy": "dummy"},
            ]
        )
        return AccessTokenConfig.from_dict(auth_config.data)



import datetime

from typing import TYPE_CHECKING, Optional, Type

from clipped.utils.tz import now

from polyaxon._config.manager import ConfigManager
from polyaxon._schemas.cli import CliConfig

if TYPE_CHECKING:
    from polyaxon._schemas.compatibility import V1Compatibility
    from polyaxon._schemas.installation import V1Installation
    from polyaxon._schemas.log_handler import V1LogHandler


class CliConfigManager(ConfigManager):
    """Manages access cli configuration .cli file."""

    VISIBILITY = ConfigManager.Visibility.GLOBAL
    CONFIG_FILE_NAME = ".cli"
    CONFIG: Type[CliConfig] = CliConfig

    @classmethod
    def reset(
        cls,
        current_version: Optional[str] = None,
        installation: Optional["V1Installation"] = None,
        compatibility: Optional["V1Compatibility"] = None,
        log_handler: Optional["V1LogHandler"] = None,
        last_check: Optional[datetime.datetime] = None,
    ) -> Optional[CliConfig]:
        if not any(
            [current_version, installation, compatibility, log_handler, last_check]
        ):
            return
        cli_config = cls.get_config_or_default()
        if current_version is not None:
            cli_config.current_version = current_version
        if installation is not None:
            cli_config.installation = installation
        if compatibility is not None:
            cli_config.compatibility = compatibility
        if log_handler is not None:
            cli_config.log_handler = log_handler
        if last_check is not None:
            cli_config.last_check = cli_config.get_last_check(last_check)

        CliConfigManager.set_config(config=cli_config)
        return cli_config

    @classmethod
    def should_check(cls, interval: Optional[int] = None) -> bool:
        config = cls.get_config_or_default()
        should_check = config.should_check(interval=interval)
        if should_check:
            cls.reset(last_check=now())
        return should_check

    @classmethod
    def get_config_from_env(cls) -> Optional[CliConfig]:
        pass



from typing import Optional, Union
from typing_extensions import Annotated, Literal

from clipped.compact.pydantic import Field, StrictFloat, StrictStr
from clipped.types.ref_or_obj import BoolOrRef, FloatOrRef, IntOrRef

from polyaxon._flow.optimization import V1Optimization
from polyaxon._schemas.base import BaseSchemaModel


class V1MedianStoppingPolicy(BaseSchemaModel):
    _IDENTIFIER = "median"
    _USE_DISCRIMINATOR = True

    kind: Literal[_IDENTIFIER] = _IDENTIFIER
    evaluation_interval: IntOrRef = Field(alias="evaluationInterval")
    min_interval: Optional[IntOrRef] = Field(alias="minInterval")
    min_samples: Optional[IntOrRef] = Field(alias="minSamples")


class V1TruncationStoppingPolicy(BaseSchemaModel):
    _IDENTIFIER = "truncation"

    kind: Literal[_IDENTIFIER] = _IDENTIFIER
    percent: FloatOrRef
    evaluation_interval: IntOrRef = Field(alias="evaluationInterval")
    min_interval: Optional[IntOrRef] = Field(alias="minInterval")
    min_samples: Optional[IntOrRef] = Field(alias="minSamples")
    include_succeeded: Optional[BoolOrRef] = Field(alias="includeSucceeded")


class V1DiffStoppingPolicy(BaseSchemaModel):
    _IDENTIFIER = "diff"

    kind: Literal[_IDENTIFIER] = _IDENTIFIER
    percent: FloatOrRef
    evaluation_interval: IntOrRef = Field(alias="evaluationInterval")
    min_interval: Optional[IntOrRef] = Field(alias="minInterval")
    min_samples: Optional[IntOrRef] = Field(alias="minSamples")


V1EarlyStoppingPolicy = Annotated[
    Union[V1MedianStoppingPolicy, V1TruncationStoppingPolicy, V1DiffStoppingPolicy],
    Field(discriminator="kind", alias="earlyStopping"),
]


class V1MetricEarlyStopping(BaseSchemaModel):
    """Metric early stopping is an early stopping strategy based on metrics of runs,
    it allows to terminate a dag, a mapping, or hyperparameter tuning when a run's metric(s)
    meet(s) one or multiple conditions.

    If no policy is set and a metric early stopping condition is met the pipeline will be marked
    as succeeded and all pending or running operations will be stopped.

    If a policy is set only the runs that validate the policy will be stopped.

    Args:
        kind: str, should be equal to `metric_early_stopping`
        metric: str
        value: float
        optimization: Union["maximize", "minimize"]
        policy: Union[V1MedianStoppingPolicy, V1TruncationStoppingPolicy, V1DiffStoppingPolicy],
             optional

    ## YAML usage

    ```yaml
    >>> earlyStopping:
    >>>   - kind: metric_early_stopping
    >>>     metric: loss
    >>>     value: 0.001
    >>>     optimization: minimize
    >>>   - kind: metric_early_stopping
    >>>     metric: accuaracy
    >>>     value: 0.9
    >>>     optimization: maximize
    ```

    ## Python usage

    ```python
    >>> from polyaxon.schemas import V1MetricEarlyStopping, V1Optimization
    >>> early_stopping = [
    >>>     V1MetricEarlyStopping(metric="loss", optimization=V1Optimization.MINIMIZE),
    >>>     V1MetricEarlyStopping(metric="accuracy", optimization=V1Optimization.MAXIMIZE),
    >>> ]
    ```

    ## Fields

    ### kind

    The kind signals to the CLI, client, and other tools
    that this early stopping is `metric_early_stopping`.

    If you are using the python client to create the early stopping,
    this field is not required and is set by default.

    ```yaml
    >>> earlyStopping:
    >>>   - kind: metric_early_stopping
    ```

    ### metric

    The metric to track for checking the early stopping condition.
    This metric should be logged using one of the tracking modules or API.

    ```yaml
    >>> earlyStopping:
    >>>   - metric: loss
    ```

    ### value

    The metric value for checking the early stopping condition.

    ```yaml
    >>> earlyStopping:
    >>>   - value: 0.5
    ```

    ### optimization

    The optimization defines the goal or how to measure the performance of the defined metric.

    ```yaml
    >>> earlyStopping:
    >>>   - optimization: maximize
    ```

    ### policy

    A policy allows to defines how to evaluate the metric value against the defined value,
    there are a couple of policies:
     * MedianStopping: Early stopping with median stopping,
         this policy computes running medians across all runs and stops
         those whose best performance is worse than the median of the running runs.
     * DiffStopping: Early stopping with diff factor stopping,
         this policy computes checked runs against the best run and
         stops those whose performance is worse than the best by the factor defined by the user.
     * TruncationStopping: Early stopping with truncation stopping,
         this policy stops a percentage of all running runs at every evaluation.
    """

    _IDENTIFIER = "metric_early_stopping"

    kind: Literal[_IDENTIFIER] = _IDENTIFIER
    metric: StrictStr
    value: FloatOrRef
    optimization: Union[StrictFloat, V1Optimization]
    policy: Optional[V1EarlyStoppingPolicy]


class V1FailureEarlyStopping(BaseSchemaModel):
    """Failure early stopping is an early stopping strategy based on statuses of runs that allows
    to terminate a dag, a mapping, or hyperparameter tuning group
    when they reach a certain level of failures.

    If a percentage of the runs in the pipeline fail,
    the pipeline will be marked as failed as well,
    and all pending or running operations will be stopped.

    Args:
        kind: str, should be equal to `failure_early_stopping`
        percent: int (>0, <=99)


    ## YAML usage

    ```yaml
    >>> earlyStopping:
    >>>   - kind: failure_early_stopping
    >>>     percent: 50
    ```

    ## Python usage

    ```python
    >>> from polyaxon.schemas import V1FailureEarlyStopping
    >>> early_stopping = [V1FailureEarlyStopping(percent=50)]
    ```

    ## Fields

    ### kind

    The kind signals to the CLI, client, and other tools
    that this early stopping is `failure_early_stopping`.

    If you are using the python client to create the early stopping,
    this field is not required and is set by default.

    ```yaml
    >>> earlyStopping:
    >>>   - kind: failure_early_stopping
    ```

    ### percent

    The percentage of failed runs at each evaluation interval,
    should be a value between 1 and 99.

    ```yaml
    >>> earlyStopping:
    >>>   - kind: failure_early_stopping
    >>>     percent: 30
    ```
    """

    _IDENTIFIER = "failure_early_stopping"

    kind: Literal[_IDENTIFIER] = _IDENTIFIER
    percent: FloatOrRef
